import subprocess
from subprocess import Popen, PIPE
from .logger import Logger


class BashHelpers(Logger):

    def __init__(self,dry_run=False):
        self.dry_run = dry_run

    def run_bash_cmd(self,cmd):
        if isinstance(cmd,str):
            cmd = cmd.split(" ")
        elif not isinstance(cmd,list):
            raise Exception("run_bash_cmd must be passed a bash command as a str or a list")
        self.logger.notice(" ".join(cmd))
        if self.dry_run:
            return None,None
        session = subprocess.Popen(cmd, stdout=PIPE, stderr=PIPE)
        stdout, stderr = (x.decode("utf-8") for x in session.communicate())

        if 'ERROR' in stderr:
            self.logger.critical(stderr)
            raise Exception("failled executing bash command")
        return stdout,stderr



import carrot
import os
import sys
import json
import glob
import inspect
import time

from . import extract
from .dag import make_dag,make_report_dag

from . import bclink_helpers 

from .file_helpers import (
    load_json,
    load_json_delta,
    load_csv,
    load_tsv,
    load_sql,
    create_csv_store,
    create_sql_store,
    create_bclink_store,
    remove_missing_sources_from_rules,
    filter_rules_by_destination_tables,
    filter_rules_by_object_names,
    get_separator_from_filename,
    get_file_map_from_dir,
    get_mapped_fields_from_rules,
    get_source_tables_from_rules,
    get_subfolders,
    get_files,
    diff_csv
)

from .rules_helpers import(
    get_person_ids,
    get_source_field,
    get_source_table,
    apply_rules,
    load_from_file
)

_DEBUG = False

def set_debug(value):
    global _DEBUG
    _DEBUG = value
    
def get_classes(format=False):
    retval = get_classes_from_tool(format=format)
    config_folder = os.environ.get('CARROT_CONFIG_FOLDER')
    if config_folder is not None:
        sys.path.append(config_folder)
        files = [x for x in os.listdir(config_folder) if x.endswith(".py") and not x.startswith('__')]
        for fname in files:
            mname = fname.split(".")[0]
            module = __import__(mname,fromlist=[fname])
            path = os.path.join(config_folder,fname)
            defined_classes = {
                fname: {
                    'module':m[1].__module__,
                    'path': path,
                    'last-modified': time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(os.path.getmtime(os.path.join(config_folder,fname))))
                }
                for m in inspect.getmembers(module, inspect.isclass)
                if m[1].__module__ == module.__name__
            }
            retval.update(defined_classes)

    return retval
    
def get_classes_from_tool(format=format):
    from carrot.cdm import classes
    _dir = os.path.dirname(classes.__file__)
    files = [x for x in os.listdir(_dir) if x.endswith(".py") and not x.startswith('__')]
    retval = {}
    for fname in files:
        path = os.path.join(_dir,fname)
        mname = fname.split(".")[0]
        mname = '.'.join([classes.__name__, mname])
        if os.path.islink(path):
            link = os.readlink(path)
            if os.path.isfile(link) == False:
                os.unlink(path)
                continue
        
        module = __import__(mname,fromlist=[fname])
        defined_classes = {
            fname: {
                'module':m[1].__module__,
                'path': path  if not os.path.islink(path) else os.readlink(path),
                'sympath': path,
                'last-modified': time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(os.path.getmtime(os.path.join(_dir,fname))))
            }
            for m in inspect.getmembers(module, inspect.isclass)
            if m[1].__module__ == module.__name__
        }
        retval.update(defined_classes)
    if format:
        return json.dumps(retval,indent=6)
    else:
        return retval




import pandas as pd
import inspect
import os
import sys
import click
import json
import yaml
import glob
import copy
import subprocess
import carrot
import carrot.tools as tools

@click.group(help="Commands for mapping data to the OMOP CommonDataModel (CDM).")
def run():
    pass


@click.command(help="Generate a python class from the OMOP mapping json")
@click.option("--name",
              help="give the name of the dataset, this will be the name of the .py class file created")
@click.option("--register",
              is_flag=True,
              help="also register the python config with the tool")
@click.argument("rules")
@click.pass_context

def make_class(ctx,name,rules,register):
    data = tools.load_json(rules)
    if name == None:
        name = data['metadata']['dataset']

    fname = tools.extract.make_class(data,name)
    if register:
        ctx.invoke(register_class,pyconfig=fname)
    return fname

@click.command(help="Register a python class with the tool")
@click.argument("pyconfig")
def register_class(pyconfig):
    pyconfig = os.path.abspath(pyconfig)
    _,conf_extension = os.path.splitext(pyconfig)
    if conf_extension != '.py':
        raise NotImplementedError(f"You're trying to register the file {pyconfig} which isnt a .py file")
    tools.extract.register_class(pyconfig)

    
    
@click.command(help="List all the python classes there are available to run")
def list_classes():
    print (json.dumps(tools.get_classes(),indent=6))

@click.command(help="remove a registered class")
@click.pass_context
@click.argument("name")
def remove_class(ctx,name):
    classes = tools.get_classes()
    if name not in classes:
        print (f"{name} is not a registered class")
        ctx.invoke(list_classes)
    else:
        _class = classes[name]
        os.unlink(_class['sympath'])

@click.command(help="Execute the running of the test dataset")
@click.pass_context
def test(ctx):
    _dir = os.path.dirname(os.path.abspath(carrot.__file__))
    _dir = f"{_dir}{os.path.sep}data{os.path.sep}test"
    inputs = glob.glob(f"{_dir}{os.path.sep}inputs{os.path.sep}*.csv")
    rules = f"{_dir}{os.path.sep}rules{os.path.sep}rules_14June2021.json"
    output_folder = 'tests'
    ctx.invoke(run,inputs=inputs,rules=rules,output_folder=output_folder)
    for fname in glob.glob(f"{output_folder}{os.path.sep}*.tsv"):
        tools.diff_csv(fname,fname)


@click.command(help="apply operations (transforms) to a dataset")
@click.option("--config","-c",required=True,type=str)
@click.option("--number-of-rows-per-chunk","--nc",default=1e5,type=int)
@click.option("--output-folder","-o",required=True,type=str)
@click.argument("inputs",nargs=-1,required=True)
def transform(inputs,config,number_of_rows_per_chunk,output_folder):

    logger = tools.logger.Logger("transform")
    
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)
    
    if os.path.isfile(config):
        config = json.load(open(config))
    else:
        config = json.loads(config)

    logger.info(json.dumps(config,indent=6))
        
    inputs = {
        os.path.basename(x):x
        for x in inputs
    }
    input_data = tools.load_csv(inputs,chunksize=number_of_rows_per_chunk)

    operation_tools = carrot.cdm.OperationTools()

    header=True
    mode='w'

    i = 0 
    while True:
        if i > 0:
            mode = 'a'
            header=False
        i+=1
        for fname,rules in config.items():
            logger.info(f"Working on {fname}") 
            df = input_data[fname]
            for colname,operations in rules.items():
                series = df[colname]
                for operation in operations:
                    logger.info(f".. transforming '{colname}' with operation '{operation}'")
                    fn_operation = operation_tools[operation]
                    series = fn_operation(series)
                df[colname] = series
            fout = f"{output_folder}/{fname}"
            df.to_csv(fout,header=header,mode=mode,index=False)

        try:
            input_data.next()
        except StopIteration:
            break

@click.command()
@click.option("--number-of-rows-per-chunk","--nc",default=1e5,type=int)
@click.option("--output-folder","-o",required=True,type=str)
@click.argument("inputs",nargs=-1,required=True)
def format(inputs,number_of_rows_per_chunk,output_folder):
    """
    Format a CDM model by passing all CDM objects and applying formatting.
    """
    types = list(set([
        os.path.splitext(fname)[1]
        for fname in inputs
    ]))
    if len(types) > 1:
        raise Exception(f"Running with mixed input files '{types}'. Only input tsv or csv files.")
    types = types[0]
    
    inputs = {
        os.path.basename(x):x
        for x in inputs
    }
    if types == '.csv':
        inputs = tools.load_csv(inputs,chunksize=number_of_rows_per_chunk)
    else:
        inputs = tools.load_tsv(inputs,chunksize=number_of_rows_per_chunk)

    cdm = carrot.cdm.CommonDataModel.from_existing(inputs=inputs,
                                                      output_folder=output_folder,
                                                      format_level=1)
    #cdm.save_files = False
    cdm.process()
    cdm.end()

@click.command()
@click.option("--output-folder","-o",required=True,type=str,help='specify an output folder of where to put the merged tables')
@click.argument("inputs",nargs=-1,required=True)
def merge(inputs,output_folder):
    """
    Build a CDM model from subfiles
    """
    types = list(set([
        os.path.splitext(fname)[1]
        for fname in inputs
    ]))
    if len(types) > 1:
        raise Exception(f"Running with mixed input files '{types}'. Only input tsv or csv files.")
    types = types[0]
    
    inputs = {
        os.path.basename(x):x
        for x in inputs
    }
    if types == '.csv':
        inputs = tools.load_csv(inputs)#,chunksize=number_of_rows_per_chunk)
    else:
        inputs = tools.load_tsv(inputs)#,chunksize=number_of_rows_per_chunk)

    outputs = carrot.tools.create_csv_store(output_folder=output_folder,sep='\t',write_mode='w',write_separate=False)
        
    cdm = carrot.cdm.CommonDataModel.from_existing(inputs=inputs,
                                                      do_mask_person_id=False,
                                                      drop_duplicates=True,
                                                      format_level=0,
                                                      outputs=outputs)
    cdm.process()


@click.command(help="loading")
@click.option("--config","-c",required=True,type=str)
@click.argument("inputs",nargs=-1,required=True)
def load(inputs,config):
    inputs = {os.path.basename(x).split('.')[0]:x for x in inputs}
    config = json.loads(config)
    if 'bclink' in config:
        outputs = carrot.io.BCLinkDataCollection(config['bclink'])
    else:
        raise NotImplementedError(f"Not setup to implement configuration for load {config}")

    for destination_table,fname in inputs.items():
        outputs.load(fname,destination_table)

    outputs.finalise()
    
@click.command()
@click.option("--input",
              required=True,
              help='input csv file')
@click.option("--column",
              required=True,
              help="which column of the input data to modify")
@click.option("--operation",
              required=True,
              help="which operation to apply to the column")
def format_input_data(column,operation,input):
    """
    Useful formatting command for applying an operation on some data before being passed into the ETL-Tool

    """
    optools = carrot.cdm.OperationTools()
    allowed_operations = optools.keys()
    if operation not in allowed_operations:
        raise Exception(f"Operation '{operation}' is not a known operation. Choose from {allowed_operations}")
    df_input = pd.read_csv(input)
    df_input[column] = optools[operation](df_input[column])
    n = 5 if len(df_input) > 5 else len(df_input)
    print (df_input[column].sample(n))
    df_input.to_csv(input)


def _condor(name,commands,jobscript):
    #hold_jid {name} - wait for previous job
    def wrapper(args):
        script = jobscript + '\n' + ' '.join(commands)
        fname = f"{name}.sh"
        with open(fname,"w") as f:
            f.write(script)
        output = subprocess.check_output(['qsub','-N',name,fname]).decode()
        return output
    return wrapper

from importlib import import_module

@click.command()
@click.option("analysis_name","--analysis",default=None,type=str)
@click.argument("analysis_file")
@click.pass_context
def analysis(ctx,analysis_file,analysis_name):
    """
    Use this command to run analyses on input data (in the CDM format) given a configuration yaml file
    """
    _dir = os.path.dirname(analysis_file)
    fname = os.path.splitext(os.path.basename(analysis_file))[0]
    sys.path.append(_dir)
    module = import_module(fname)
    clsmembers = inspect.getmembers(module, inspect.isclass)
    name,cls = clsmembers[0]
    obj = cls()
    if analysis_name:
        ana = obj.get_analysis(analysis_name)
        res = obj.run_analysis(ana)
        print (res)
    else:
        #print (obj.get_analyses())
        #exit(0)
        obj.run_analyses()


@click.command()
@click.option("--max-workers",default=None,type=int)
@click.option("--batch",default=None,type=click.Choice(['condor']))
@click.option("analysis_names","--analysis-name",default=None,multiple=True,type=str)
@click.argument("config")
@click.pass_context
def ___analysis(ctx,config,analysis_names,max_workers,batch):
    """
    Use this command to run analyses on input data (in the CDM format) given a configuration yaml file
    """    
    from importlib import import_module
    fname = config
    stream = open(config) 
    config = yaml.safe_load(stream)
    
    inputs = config['cdm']
    inputs = carrot.tools.load_tsv(config['cdm'],
                                      dtype=None)

    cdm = carrot.cdm.CommonDataModel.load(inputs=inputs)

    analyses = config['analyses']

    jobscript = config['condor']['jobscript']
    

    if analysis_names:
        temp = copy.copy(analyses)
        for name in temp:
            if name not in analysis_names:
                analyses.pop(name)

    for name,_def in analyses.items():
        analysis = _def['analysis']
        cohort = _def['filter']
        if batch:
            if batch == 'condor':
                commands = copy.copy(sys.argv)
                commands.remove('--batch')
                commands.remove('condor')
                commands.extend(['--analysis-name',name])
                
                f = _condor(name=name,commands=commands,jobscript=jobscript)
            else:
                raise NotImplementedError(f"{batch} mode for --batch not a thing")
        else:
            func = import_module(analysis)
            f = func.create_analysis(cohort)
        cdm.add_analysis(f)
    
    results = cdm.run_analyses(max_workers=max_workers)

    
        
@click.command()
@click.option("--rules",
              required=True,
              help="input json file containing all the mapping rules to be applied")
@click.option("--indexing-conf",
              default=None,
              help="configuration file to specify how to start the indexing")
@click.option("--csv-separator",
              default='\t',
              type=click.Choice([';',':','\t',',',' ',]),
              help="choose a separator to use when dumping output csv files")
@click.option("--use-profiler",
              is_flag=True,
              help="turn on saving statistics for profiling CPU and memory usage")
@click.option("format_level","--format-level",
              default='1',
              type=click.Choice(['0','1','2']),
              help="Choose the level of formatting to apply on the output data. 0 - no formatting. 1 - automatic formatting. 2 (default) - check formatting (will crash if input data is not already formatted).")
@click.option("--output-folder",
              default=None,
              help="define the output folder where to dump csv files to")
@click.option("--write-mode",
              default='w',
              type=click.Choice(['w','a']),
              help="force the write-mode on existing files")
@click.option("--split-outputs",
              is_flag=True,
              help="force the output files to be split into separate files")
@click.option("--allow-missing-data",
              is_flag=True,
              help="don't crash if there is data tables in rules file that hasnt been loaded")
@click.option("output_database","--database",
              default=None,
              help="define the output database where to insert data into")
@click.option("-nc","--number-of-rows-per-chunk",
              default=-1,
              help="Choose the number of rows (INTEGER) of input data to load (chunksize). The option 'auto' will work out the ideal chunksize. Inputing a value <=0 will turn off data chunking (default behaviour).")
@click.option("-np","--number-of-rows-to-process",
              default=None,
              type=int,
              help="the total number of rows to process")
@click.option("--person-id-map",
              default=None,
              help="pass the location of a file containing existing masked person_ids")
@click.option("--db",
              default=None,
              help="instead, pass a connection string to a db")
@click.option("--merge-output",
              is_flag=True,
              help="merge the output into one file")
@click.option("no_mask_person_id","--parse-original-person-id",
              is_flag=True,
              help="turn off automatic conversion (creation) of person_id to (as) Integer")
@click.option("dont_automatically_fill_missing_columns","--no-fill-missing-columns",
              is_flag=True,
              help="Turn off automatically filling missing CDM columns")
@click.option("log_file","--log-file",
              default = 'none',
              help="specify a path for a log file")
@click.option("--max-rules",
              default = None,
              type=int,
              help="maximum number of rules to process")
@click.option("objects","--object",
              default = None,
              multiple=True,
              type=str,
              help="give a list of objects by name to process")
@click.option("tables","--table",
              default = None,
              multiple=True,
              type=str,
              help="give a list of tables by name to process")
@click.argument("inputs",
                required=False,
                nargs=-1)
@click.pass_context
def map(ctx,rules,inputs,format_level,
        output_folder,output_database,
        csv_separator,use_profiler,log_file,
        no_mask_person_id,indexing_conf,
        person_id_map,max_rules,merge_output,
        objects,tables,db,write_mode,split_outputs,
        dont_automatically_fill_missing_columns,
        number_of_rows_per_chunk,allow_missing_data,
        number_of_rows_to_process):
    """
    Perform OMOP Mapping given an json file and a series of input files

    INPUTS should be a space separated list of individual input files or directories (which contain .csv files)
    """

    if output_folder is None:
        output_folder = f'{os.getcwd()}{os.path.sep}output_data{os.path.sep}'

    #if log_file == 'auto' and carrot.params['log_file'] is None:
    if log_file == 'auto':
        log_file = f"{output_folder}{os.path.sep}logs{os.path.sep}carrot.log"
        carrot.params['log_file'] = log_file
    elif log_file == 'none':
        pass
    else:
        carrot.params['log_file'] = log_file
        
    #load the json loads
    if type(rules) == dict:
        config = rules
    else:
        config = tools.load_json(rules)

    if tables:
        tables = list(set(tables))
        config = carrot.tools.filter_rules_by_destination_tables(config,tables)

    if objects:
        objects = list(set(objects))
        config = carrot.tools.filter_rules_by_object_names(config,objects)
        
    if max_rules:
        i = 0
        n = max_rules
        new = {}
        for destination_table,rule_set in config['cdm'].items():
            if destination_table == 'person':
                new[destination_table] = rule_set
            else:
                for name,_rules in rule_set.items():
                    if i>=n:
                        break
                    if destination_table not in new:
                        new[destination_table] = {}
                    new[destination_table][name] = _rules
                    i+=1
            
        config['cdm'] = new

    name = config['metadata']['dataset']

    if indexing_conf is not None:
        if isinstance(indexing_conf,dict):
            pass
        elif indexing_conf.endswith(".json") and os.path.exists(indexing_conf):
            indexing_conf = tools.load_json(indexing_conf)
        elif indexing_conf.endswith(".csv") and os.path.exists(indexing_conf):
            try:
                indexing_conf = pd.read_csv(indexing_conf,header=None,index_col=0)[1].to_dict()
            except pd.errors.EmptyDataError:
                indexing_conf = None
                pass
                
    
    #automatically calculate the ideal chunksize
    if number_of_rows_per_chunk == 'auto':
        #get the fields that are going to be used/loaded
        used_fields = tools.get_mapped_fields_from_rules(config)
        #calculate the number of fields that are to be used per dataset
        n_used_fields = [ len(sublist) for sublist in used_fields.values() ]
        #find what's the largest number of fields loaded by any dataset
        max_n_used_fields = max(n_used_fields)
        #get the number of files used
        n_files = len(n_used_fields)
        
        # If there is one dataset and one column being used, the max loaded to memory
        #   is 2million rows (this is fairly arbitrary)
        #   it is an approximation assuming the data in the values is relatively small
        #   this should keep the memory usage down
        # When there is more fields and more files loaded, reduce the of rows per chunk
        max_n_rows = 2e6
        number_of_rows_per_chunk = int(max_n_rows/(max_n_used_fields*n_files))
    else:
        try:
            number_of_rows_per_chunk = int(number_of_rows_per_chunk)
        except ValueError:
            raise ValueError(f"number_of_rows_per_chunk must be an Integer or 'auto', you inputted '{number_of_rows_per_chunk}'")
        
        #turn off chunking if 0 or negative chunksizes are given
        if number_of_rows_per_chunk <= 0 :
            number_of_rows_per_chunk = None
    
    #check if exists
    if any('*' in x for x in inputs):
        data_dir = os.path.dirname(carrot.__file__)
        data_dir = f'{data_dir}{os.path.sep}data{os.path.sep}'

        new_inputs = []
        for i,x in enumerate(inputs):
            if not os.path.exists(x):
                new_inputs.extend(glob.glob(f"{data_dir}{os.path.sep}{x}"))
            else:
                new_inputs.append(x)
        inputs = new_inputs

    inputs = list(inputs)
    
    for x in inputs:
        if os.path.isdir(x):
            inputs.remove(x)
            inputs.extend(glob.glob(f'{x}{os.path.sep}*.csv'))
        
    #convert the list into a map between the filename and the full path
    inputs = {
        os.path.basename(x):x
        for x in inputs
    }

    if db:
        inputs = tools.load_sql(connection_string=db,chunksize=number_of_rows_per_chunk,nrows=number_of_rows_to_process)
    else:
        if allow_missing_data:
            config = carrot.tools.remove_missing_sources_from_rules(config,inputs)

        inputs = tools.load_csv(inputs,
                                rules=config,
                                chunksize=number_of_rows_per_chunk,
                                nrows=number_of_rows_to_process)

    #do something with
    #person_id_map
        
    if isinstance(output_database,dict):
        if 'bclink' in output_database:
            outputs = carrot.tools.create_bclink_store(bclink_settings=output_database['bclink'],
                                                          output_folder=output_database['cache'],
                                                          sep=csv_separator,
                                                          write_separate=split_outputs,
                                                          write_mode=write_mode)
        else:
            raise NotImplementedError(f"dont know how to configure outputs... {output_database}")   
    elif output_database == None:
        outputs = carrot.tools.create_csv_store(output_folder=output_folder,
                                                   sep=csv_separator,
                                                   write_separate=split_outputs,
                                                   write_mode=write_mode)
    else:
        outputs = carrot.tools.create_sql_store()

    #build an object to store the cdm
    cdm = carrot.cdm.CommonDataModel(name=name,
                                        inputs=inputs,
                                        format_level=format_level,
                                        do_mask_person_id=not no_mask_person_id,
                                        outputs = outputs,
                                        #output_folder=output_folder,
                                        #output_database=output_database,
                                        automatically_fill_missing_columns=not dont_automatically_fill_missing_columns,
                                        use_profiler=use_profiler)
    #allow the csv separator to be changed
    #the default is tab (\t) separation
    #if not csv_separator is None:
    #    cdm.set_csv_separator(csv_separator)
    cdm.create_and_add_objects(config)

    cdm.process(conserve_memory=True)
    cdm.close()

    if merge_output:
        ctx.invoke(merge,
                   inputs=glob.glob(f"{output_folder}{os.path.sep}*"),
                   output_folder=output_folder)
    
@click.command(help="Perform OMOP Mapping given a python configuration file.")
@click.option("--rules",
              help="input json file containing all the mapping rules to be applied")
@click.option("--pyconf","--conf",
              required=True,
              help="Run with a python configuration file")
@click.option("--object","objects",
              multiple=True,
              help="Only run specific objects.")
@click.option("--type",
              default='csv',
              type=click.Choice(['csv']),
              help="specify the type of inputs, the default is .csv inputs")
@click.option("--use-profiler",
              is_flag=True,
              help="turn on saving statistics for profiling CPU and memory usage")
@click.option("--output-folder",
              default=None,
              help="define the output folder where to dump csv files to")
@click.option("-nc","--number-of-rows-per-chunk",
              default=None,
              type=int,
              help="choose to chunk running the data into nrows")
@click.option("-np","--number-of-rows-to-process",
              default=None,
              type=int,
              help="the total number of rows to process")
@click.argument("inputs",
                #help="give a list of input files to process, and/or an input directory",
                nargs=-1)
@click.pass_context
def run_pyconfig(ctx,rules,pyconf,inputs,objects,
        output_folder,type,use_profiler,
        number_of_rows_per_chunk,
        number_of_rows_to_process):

    object_list = list(objects)
    if len(object_list) == 0:
        object_list = None

    if type != 'csv':
        raise NotImplementedError("Can only handle inputs that are .csv so far")
        
    #check if exists
    if any('*' in x for x in inputs):
        data_dir = os.path.dirname(carrot.__file__)
        data_dir = f'{data_dir}{os.path.sep}data{os.path.sep}'

        new_inputs = []
        for i,x in enumerate(inputs):
            if not os.path.exists(x):
                new_inputs.extend(glob.glob(f"{data_dir}{os.path.sep}{x}"))
            else:
                new_inputs.append(x)
        inputs = new_inputs

    inputs = list(inputs)
    for x in inputs:
        if os.path.isdir(x):
            inputs.remove(x)
            inputs.extend(glob.glob(f'{x}{os.path.sep}*.csv'))
        
    #convert the list into a map between the filename and the full path
    inputs = {
        os.path.basename(x):x
        for x in inputs
    }
    
    if output_folder is None:
        output_folder = f'{os.getcwd()}{os.path.sep}output_data{os.path.sep}'
        
    outputs = carrot.tools.create_csv_store(output_folder=output_folder)


    if not inputs:
        raise Exception('no inputs defined!')

    inputs = tools.load_csv(inputs,
                            rules=rules,
                            chunksize=number_of_rows_per_chunk,
                            nrows=number_of_rows_to_process)

    available_classes = tools.get_classes()
    if pyconf not in available_classes:
        ctx.invoke(list_classes)
        raise KeyError(f"cannot find config {pyconf}. Run 'carrot map py list' to see available classes.")
    
    module = __import__(available_classes[pyconf]['module'],fromlist=[pyconf])
    defined_classes = [
        m[0]
        for m in inspect.getmembers(module, inspect.isclass)
        if m[1].__module__ == module.__name__
    ]
    #should only be running one class anyway
    defined_class = defined_classes[0]
    cls = getattr(module,defined_class)

    #build a class object
    cdm = cls(inputs=inputs,
              outputs=outputs,
              use_profiler=use_profiler)
    #run it
    cdm.process(object_list)
      
@click.command(help="run as a Graphical User Interface (GUI)")
@click.pass_context
def gui(ctx):
    import PySimpleGUI as sg

    carrot_theme = {'BACKGROUND': 'white',
                       'TEXT': '#000000',
                       'INPUT': '#c4c6e2',
                       'TEXT_INPUT': '#000000',
                       'SCROLL': '#c4c6e2',
                       'BUTTON': ('white', '#475da7','#3DB28C'),
                       'PROGRESS': ('#01826B', '#D0D0D0'),
                       'BORDER': 1,
                       'SLIDER_DEPTH': 1,
                       'PROGRESS_DEPTH': 0}
    
    # Add your dictionary to the PySimpleGUI themes
    sg.theme_add_new('carrot', carrot_theme)
    sg.theme('carrot')

    _dir = os.path.dirname(os.path.abspath(carrot.__file__))
    data_dir = f"{_dir}{os.path.sep}data{os.path.sep}"
    
    layout = [
        [sg.Image(f'{data_dir}logo.png'),sg.T("CO-CONNECT: Dataset2CDM",font = ("Roboto", 25))],
        [sg.T('Select the rules json:')],
        [sg.Input(key='_RULES_'), sg.FilesBrowse(initial_folder=os.getcwd())],
        [sg.T('Select the input CSVs:')],
        [sg.Input(key='_INPUTS_'), sg.FilesBrowse(initial_folder=os.getcwd())],
        [sg.T('Select an output folder:')],
        [sg.Input(key='_OUTPUT_',default_text='.'), sg.FolderBrowse(initial_folder=os.getcwd())],
        #[sg.Checkbox("Mask the person_id",key="_MASK_PERSON_ID_",default=False)],
        #[[sg.T('Change the default data chunksize:'),
        #  sg.Slider(range=(0,1000000),
        #            default_value=100000,
        #            resolution=10000,
        #            orientation='horizontal')]],
        [sg.OK('Run'), sg.Cancel(button_color=('white','#3DB28C'))]
    ]

    font = ("Roboto", 15)
    
    window = sg.Window('carrot', layout, font=font)
    while True:
        event, values = window.Read()
        
        if event == 'Cancel' or event == None:
            break

        output_folder = values['_OUTPUT_']
        if output_folder == '':
            output_folder = None
            
        rules = values['_RULES_']
        if rules == '':
            sg.Popup(f'Error: please select a rules file')
            continue
        elif len(rules.split(';'))>1:
            sg.Popup(f'Error: only select one file for the rules!')
            continue

        inputs = values['_INPUTS_']
        if inputs == '':
            sg.Popup(f'Error: please select at least one file or directory for the inputs')
            continue
        inputs = inputs.split(';')

        #mask_person_id = values['_MASK_PERSON_ID_']

        try:
            #ctx.invoke(run,rules=rules,inputs=inputs,output_folder=output_folder,mask_person_id=mask_person_id)
            ctx.invoke(run,rules=rules,inputs=inputs,output_folder=output_folder)
            sg.Popup("Done!")
        except Exception as err:
            sg.popup_error("An exception occurred!",err)
            
        break
        
    window.close()
    

@click.group(help="Commands for using python configurations to run the ETL transformation.")
def py():
    pass



py.add_command(make_class,"make")
py.add_command(register_class,"register")
py.add_command(list_classes,"list")
py.add_command(remove_class,"remove")
py.add_command(run_pyconfig,"map")
run.add_command(py,"py")
run.add_command(map,"map")
run.add_command(load,"load")
run.add_command(analysis,"analysis")
run.add_command(format,"format")
run.add_command(merge,"merge")
run.add_command(transform,"transform")
run.add_command(gui,"gui")
run.add_command(test,"test")



import pandas as pd
from ...common import DestinationTable, DestinationField

class DrugExposure(DestinationTable):
    """
    CDM Drug Exposure object class
    """
    
    name = 'drug_exposure'
    def __init__(self,name=None):
        self.drug_exposure_id             = DestinationField(dtype="Integer"   , required=True , pk=True)
        self.person_id                    = DestinationField(dtype="Integer"   , required=True )
        self.drug_concept_id              = DestinationField(dtype="Integer"   , required=True )
        self.drug_exposure_start_date     = DestinationField(dtype="Date"      , required=False )
        self.drug_exposure_start_datetime = DestinationField(dtype="Timestamp" , required=True )
        self.drug_exposure_end_date       = DestinationField(dtype="Date"      , required=False )
        self.drug_exposure_end_datetime   = DestinationField(dtype="Timestamp" , required=False )
        self.verbatim_end_date            = DestinationField(dtype="Date"      , required=False )
        self.drug_type_concept_id         = DestinationField(dtype="Integer"   , required=False )
        self.stop_reason                  = DestinationField(dtype="Text20"    , required=False )
        self.refills                      = DestinationField(dtype="Integer"   , required=False )
        self.quantity                     = DestinationField(dtype="Float"     , required=False )
        self.days_supply                  = DestinationField(dtype="Integer"   , required=False )
        self.sig                          = DestinationField(dtype="Integer"   , required=False )
        self.route_concept_id             = DestinationField(dtype="Integer"   , required=False )
        self.lot_number                   = DestinationField(dtype="Text50"    , required=False )
        self.provider_id                  = DestinationField(dtype="Integer"   , required=False )
        self.visit_occurrence_id          = DestinationField(dtype="Integer"   , required=False )
        self.drug_source_value            = DestinationField(dtype="Text50"    , required=False )
        self.drug_source_concept_id       = DestinationField(dtype="Integer"   , required=False )
        self.route_source_value           = DestinationField(dtype="Text50"    , required=False )
        self.dose_unit_source_value       = DestinationField(dtype="Text50"    , required=False )
        
        if name is None:
            name = hex(id(self))
        super().__init__(name,self.name)

        
    def get_df(self,**kwargs):
        """
        Overload/append the creation of the dataframe, specifically for the drug_exposure objects
        * drug_concept_id  is required to be not null
          this can happen when spawning multiple rows from a person
          we just want to keep the ones that have actually been filled
        
        Returns:
           pandas.Dataframe: output dataframe
        """

        df = super().get_df(**kwargs)
        if self.automatically_fill_missing_columns == True:
            if df['drug_exposure_start_date'].isnull().all():
                df['drug_exposure_start_date'] = self.tools.get_date(df['drug_exposure_start_datetime'])

            if df['drug_exposure_end_date'].isnull().all():
                df['drug_exposure_end_date'] = self.tools.get_date(df['drug_exposure_end_datetime'])
        return df
    



import os
import click
import pandas
import json
import carrot.tools as tools
import carrot
import pandas as pd
import numpy as np
import datetime

@click.group(help='Commands for displaying various types of data and files.')
def display():
    pass

@click.group(help='Commands for displaying various help for the CDM')
def cdm():
    pass


@click.command(help="Display a table for the CDM")
@click.argument('names',required=True,nargs=-1)
@click.option('--markdown',is_flag=True)
@click.option('--latex',is_flag=True)
def table(names,markdown,latex):
    for name in names:
        obj = carrot.cdm.get_cdm_class(name)()
        data = []
        for field in obj.fields:
            pk = obj[field].pk
            if pk == True:
                pk = '✔'
            else:
                pk = ' '
            
            data.append([field,pk,obj[field].dtype])

        df = pd.DataFrame(data, columns=['name','pk','dtype'])#.set_index('name')
        df['table'] = name
        df.set_index('table',inplace=True)
        df.set_index('name',inplace=True)
        
        if markdown:
            df = df.to_markdown()
        elif latex:
            df = df.to_latex()
            
        print (df)

@click.command(help="Display a dataframe")
@click.argument('fname')
@click.option('--drop-col',multiple=True)
@click.option('--drop-na',is_flag=True)
@click.option('--markdown',is_flag=True)
@click.option('--latex',is_flag=True)
@click.option('--head',type=int,default=None)
@click.option('--sample',type=int,default=None)
@click.option('--separator','--sep',type=str,default=None)
def dataframe(fname,drop_na,drop_col,markdown,head,sample,separator,latex):

    #if separator not specified, get it from the file extension 
    if separator == None:
        separator = tools.get_separator_from_filename(fname)


    if sample and head:
        head = None
    df = pandas.read_csv(fname,nrows=head,sep=separator)
    if sample:
        df = df.sample(sample)
    df.set_index(df.columns[0],inplace=True)
    if drop_na:
        df = df.dropna(axis=1,how='all')

    drop_col = list(drop_col)
    df = df.drop(drop_col,axis=1)
        
    if markdown:
        df = df.to_markdown()
    elif latex:
        df = df.to_latex()

    print (df)

@click.command(help="plot from a csv file")
@click.argument('fnames',nargs=-1)
@click.option('-y',required=True,multiple=True)
@click.option('-x',required=True)
@click.option('--nbins',default=None,help='bin and take the average',type=int)
@click.option('--save-plot',default=None,help='choose the name of file to save the plot to')
def plot(fnames,x,y,save_plot,nbins):
    import matplotlib.pyplot as plt
    fig,ax = plt.subplots(len(y),figsize=(14,7))

    dfs = {
        fname:pandas.read_csv(fname)
        for fname in fnames
    }
    
    for i,_y in enumerate(y):
        ax[i].set_ylabel(_y)
        for fname in fnames:
            if not nbins:
                dfs[fname].plot(x=x,y=_y,ax=ax[i],label=fname)
            else:
                dx = dfs[fname][x]
                dy = dfs[fname][_y]
                sums, edges = np.histogram(dx, bins=nbins, weights=dy)
                print (sums)
                counts, _ = np.histogram(dx, bins=nbins)
                counts = sums / counts
                ax[i].hist(edges[1:],weights=counts,histtype='step')
                pass
    plt.show()
    if save_plot:
        fig.savefig(save_plot)


@click.command(help="Display the OMOP mapping json as a DAG")
@click.option('--orientation',default='RL',type=click.Choice(['RL','LR','TB','BT']))
@click.option('--show-concepts',is_flag=True,help="also show the concepts")
@click.option('--output-file',default='dag.gv',help="output file name")
@click.option('--format',default='pdf',help="file extension")
@click.argument("rules")
def dag(rules,orientation,show_concepts,output_file,format):
    data = tools.load_json(rules)
    if 'cdm' in data:
        data = data['cdm']
    tools.make_dag(data,output_file=output_file,format=format,orientation=orientation,render=True,show_concepts=show_concepts)

@click.command(help="Display the report json as a DAG")
@click.option('--orientation',default='RL',type=click.Choice(['RL','LR','TB','BT']))
@click.argument("f")
def report_dag(f,orientation):
    data = tools.load_json(f)
    tools.make_report_dag(data,orientation=orientation,render=True)


@click.command(help="Show the OMOP mapping json")
@click.argument("rules")
@click.option('--list-tables',is_flag=True)
@click.option('--list-fields',is_flag=True)
@click.option('operations','--add-operation',default=None,type=str)
def print_json(rules,list_fields,list_tables,operations):
    data = tools.load_json(rules)
    if list_fields or list_tables:
        data = tools.get_mapped_fields_from_rules(data)
        if list_tables:
            data = list(data.keys())

    if operations:
        operations = tools.load_json(operations)
        for field_name,operation in operations.items():
            print (field_name,operation)
        
    print (json.dumps(data,indent=6))


@click.command(help="Detect differences in either inputs or output csv files")
@click.option('--separator','--sep',type=str,default=None)
@click.option('--max-rows','-n',type=int,default=None)
@click.argument("file1")
@click.argument("file2")
def diff(file1,file2,separator,max_rows):
    tools.diff_csv(file1,file2,separator=separator,nrows=max_rows)

@click.command(help="display a delta of two rules files")
@click.argument("rules",nargs=2)
def delta(rules):
    r1,r2 = rules
    delta = carrot.tools.load_json_delta(r2,r1)
    dt = str(datetime.datetime.now())
    delta['metadata']['date_created'] = dt
    click.echo(json.dumps(delta,indent=6))
    
@click.command(help="flattern a rules json file")
@click.argument("rules")
def flatten(rules):
    data = tools.load_json(rules)
    objects = data['cdm']
    for destination_table,rule_set in objects.items():
        if len(rule_set) < 2: continue
        #print (rule_set)
        df = pd.DataFrame.from_records(rule_set).T

        #for name in df.index:
        #    print (name,len(df.loc[name]))

        df = df.loc['condition_concept_id'].apply(pd.Series)


        def merge(s):
            if s == 'term_mapping':
                print ('hiya')
                return {k:v for a in s for k,v in a.items()}
        
        print (df.groupby('source_field').agg(merge))
            
        #print (df.iloc[1])
        #print (df.iloc[1][1])
        #print (df)
        #print (df.iloc[0])
        #print (df.iloc[0].name)
        #print (df.iloc[0].apply(pd.Series))
        #print (df.iloc[1].apply(pd.Series)['term_mapping'].apply(pd.Series))
        
        exit(0)


cdm.add_command(table,"table")
display.add_command(cdm,"cdm")

display.add_command(dataframe,"dataframe")
display.add_command(plot,"plot")
display.add_command(diff,"diff")

@click.group(help='Commands for displaying json rules in various ways.')
def rules():
    pass


rules.add_command(delta,"delta")
rules.add_command(dag,"dag")
rules.add_command(print_json,"json")
rules.add_command(flatten,"flatten")
display.add_command(rules,"rules")


@click.group(help='Commands for displaying json report in various ways.')
def report():
    pass


report.add_command(print_json,"json")
report.add_command(report_dag,"dag")


display.add_command(report,"report")




from .objects import (
    Person,
    ConditionOccurrence,
    VisitOccurrence,
    Measurement,
    Observation,
    DrugExposure
)
import copy
import sys
import time

import subprocess

class analysis(object):
    def __init__(self, method):
        self._method = method
        self._name = method.__name__
    def __call__(self, obj, *args, **kwargs):
        return self._method(obj, *args, **kwargs)

       
class qsub(object):

    def __init__(self,jobscript,**kwargs):
        self.default_kwargs = kwargs
        self.jobscript = jobscript

    def set_jobscript(self,jobscript):
        self.jobscript = jobscript
    
    def __call__(self, analysis, *args, **kwargs):
        return self._qsub(analysis, *args, **kwargs)

    def make_jobscript(self,**kwargs):
        return self.jobscript.format(**kwargs)
    
    @classmethod
    def eddie(cls,**kwargs):
        default_kwargs={'runtime':"00:30:00",'memory':"1G",'anaconda':"5.3.1"}
        default_kwargs.update(kwargs)
        jobscript = r'''
#!/bin/sh
# Grid Engine options (lines prefixed with #$)
#$ -N {job_name}             
#$ -cwd                  
#$ -l h_rt={runtime}
#$ -l h_vmem={memory}

# Initialise the environment modules
. /etc/profile.d/modules.sh

# Load Python
module load anaconda/{anaconda}

# Load the virtual environment
source /exports/applications/apps/SL7/anaconda/5.3.1/etc/profile.d/conda.sh
conda activate carrot-env

{command}
'''
        obj = cls(jobscript=jobscript,**default_kwargs)
        
        return obj

    def run(self,commands,jobname="qsub_job"):
        kwargs = self.default_kwargs
        kwargs.update({'job_name':jobname,'command':" ".join(commands)})
        jobscript = self.make_jobscript(**kwargs)
        fname = f"{jobname}.sh"
        with open(fname,"w") as f:
            f.write(jobscript)
        output = subprocess.check_output(['qsub','-N',jobname,fname]).decode()
        return output

    def _qsub(self,analysis,*args,**kwargs):        
        def wrapper(model,*args,**kwargs):
            commands = copy.copy(sys.argv)
            if "--analysis" in commands:
                return analysis(model,*args,**kwargs)
            else:
                commands.extend(["--analysis",analysis.__name__])
                name = "analysis"
                _id = format(id(analysis),'X')
                return self.run(commands=commands,jobname=f"{name}_{analysis.__name__}_{_id}")
        return wrapper
    #return _qsub

def load_file(_input):
    def func(self):
        for colname in _input:
            self[colname].series = _input[colname]
    return func


def from_table(obj,table):
    df = obj.inputs[table]
    for colname in df.columns:
        obj[colname].series = df[colname]
    return obj

def define_table(cls):
    def decorator(defs):
        obj = cls()
        obj.define = defs
        obj.set_name(defs.__name__)
        return obj
    return decorator

def define_person(defs):
    c = Person()
    c.define = defs
    c.set_name(defs.__name__)
    return c

def define_condition_occurrence(defs):
    c = ConditionOccurrence()
    c.define = defs
    c.set_name(defs.__name__)
    return c

def define_visit_occurrence(defs):
    c = VisitOccurrence()
    c.define = defs
    c.set_name(defs.__name__)
    return c

def define_measurement(defs):
    c = Measurement()
    c.define = defs
    c.set_name(defs.__name__)
    return c

def define_observation(defs):
    c = Observation()
    c.define = defs
    c.set_name(defs.__name__)
    return c

def define_drug_exposure(defs):
    c = DrugExposure()
    c.define = defs
    c.set_name(defs.__name__)
    return c




import scipy.stats
import time
import numpy as np
import json

def create_analysis(_filter):
    def ana(model):
        df = model.filter(_filter,dropna=True)
        if len(df) == 0:
            return {'fit':None,'hist':None}

        df['age'] = 2022 - df['year_of_birth']
        res = scipy.stats.linregress(x=df['age'],y=df['value_as_number'])
        attributes = ['intercept', 'intercept_stderr', 'pvalue', 'rvalue', 'slope', 'stderr']
        res = {a:getattr(res,a) for a in attributes}
        res.update({'n':len(df)})
        antibodies = df['value_as_number']
        hist = np.histogram(antibodies,bins=20, range=(0,300))
        
        #time.sleep(10)
        return {'fit':res,'hist':hist}

        # #time.sleep(20)
        # fname = 'res_'+hex(id(_filter))+'.json'
        # with open(fname,"w") as f:
        #     json.dump(res,f,indent=6)
        # return res

    return ana



from jinja2 import Template

cls = Template(r'''from carrot.cdm import define_person, define_condition_occurrence, define_visit_occurrence, define_measurement, define_observation, define_drug_exposure
from carrot.cdm import CommonDataModel
import json

class {{ name }}(CommonDataModel):
    {{ init }}
    {% for object in objects -%}
    {{ object }}
    {%- endfor %}

if __name__ == '__main__':
    {{ name }}()

''')

init = Template(r'''
    def __init__(self,**kwargs):
        """ 
        initialise the inputs and setup indexing 
        """
        super().__init__(**kwargs)
        {% if person_ids %}
        #set primary key indexing so tables can be linked
        self.set_indexing({{ person_ids }})
        {% endif %}
''')

'''
{% for ds,pk in person_ids.items() -%}
        self.inputs["{{ ds }}"].index = self.inputs["{{ ds }}"]["{{ pk }}"].rename('index')
        {% endfor %} 
'''

rule = Template(r'''self.{{ destination_field }}.series = self.inputs["{{ source_table }}"]["{{ source_field }}"]''')
operation = Template(r'''self.{{ destination_field }}.series = self.tools.{{ operation }}(self.{{ destination_field }}.series)''')

obj = Template(r'''
    @define_{{ object_name }}
    def {{ function_name }}(self):
        """
        Create CDM object for {{ object_name }}
        """
        {% for rule in map_rules -%}
        {{ rule }}
        {% endfor %}
''')




import os
import json
import re
import pandas as pd
import numpy as np
import collections
from enum import Enum
from carrot.cdm.operations import OperationTools
from carrot.tools.logger import Logger

class RequiredFieldIsNone(Exception):
    pass

class ConvertDataType(Exception):
    pass

class FailedRequiredCheck(Exception):
    pass

class FormattingError(Exception):
    pass

class DataStandardError(Exception):
    pass

class BadInputs(Exception):
    pass


class FormatterLevel(Enum):
    OFF = 0
    ON  = 1
    CHECK = 2

class DataFormatter(collections.OrderedDict,Logger):
    """
    Class for formatting DestinationFields in the CommonDataModel

    Inherits from an ordered dictionary, and maps datatypes to lambda functions.
    The lamba functions encode how to transform and format a pandas series given the datatype.

    """

    def check_formatting(self,series,function,nsample=50,tolerance=0.3):
        """
        Apply a formatting function to a subset of a series
        Args:
            series (pandas.Series) : input data series
            function (built-in function): formatting function to be applied
            nsample (int): number of rows to sample to make checks on (default = 50)
        Returns:
           series : modified or original pandas.Series object

        """
        # get the number of rows of the datframe
        n = len(series)
        nsample = nsample if n > nsample else n


        #sample the series
        series_slice = series.sample(nsample)
        #format the sample of the series
        series_slice_formatted = function(series_slice)

        #if it's just formatting of a number, just return the series if no error has been raised
        if series_slice_formatted.dtype == 'Float64':
            return series

        #if it's formatting of text i.e. date string 
        #and the pre- and post-formatting of the series are equal
        #dont waste time formatting the entire series, just return it as it is
        series_slice_values = series_slice.dropna().astype(str).unique()
        series_slice_formatted_values = series_slice_formatted.dropna().astype(str).replace('', np.nan).dropna().unique()
        
        if np.array_equal(series_slice_values,series_slice_formatted_values):
            self.logger.debug(f'Sampling {nsample}/{n} values suggests the column '\
                              f'{series.name}" is  already formatted!!')
            return series
        else:
            a=np.array(series_slice.values,dtype=str)
            b=np.array(series_slice_formatted.values,dtype=str)
            
            are_equal = a==b
            ngood = are_equal.sum()
            fraction_good = round(ngood / nsample,2)
            
            logger = self.logger.critical if fraction_good <= tolerance else self.logger.warning
            
            logger(f'Tested fomatting {nsample} rows of {series.name}. The original data is not in the right format.')

            df_bad = pd.concat([series_slice[~are_equal],series_slice_formatted[~are_equal]],axis=1)
            df_bad.columns = ['original','should be']
            
            self.logger.warning(f"\n {df_bad}")

            if logger == self.logger.critical:
                logger(f"Fraction of good columns = {fraction_good} ({ngood} / {nsample} ), is below the tolerance threshold={tolerance}")
                raise DataStandardError(f"{series.name} has not been formatted correctly")
            else:
                logger(f"Fraction of good columns ={fraction_good} ({ngood} / {nsample} ), is above the tolerance threshold={tolerance}")

    
    def __init__(self,errors='coerce'):
        super().__init__()

        self['Integer'] = lambda x : pd.to_numeric(x,errors=errors).astype('Int64')
        self['Float']   = lambda x : pd.to_numeric(x,errors=errors).astype('Float64')
        self['Text20']  = lambda x : x.fillna('').astype(str).apply(lambda x: x[:20])
        self['Text50']  = lambda x : x.fillna('').astype(str).apply(lambda x: x[:50])
        self['Text60']  = lambda x : x.fillna('').astype(str).apply(lambda x: x[:60])

        self['Timestamp'] = lambda x : pd.to_datetime(x,errors=errors)\
                                        .dt.strftime('%Y-%m-%d %H:%M:%S.%f')
        self['Date'] = lambda x : pd.to_datetime(x,errors=errors).dt.date


class DestinationField(object):
    """
    CommonDataModel Table Destination Field.

    Object for handling output columns (destination fields) in a 
    Destination Table

    Attributes:
       series (pandas.Series): raw column data in the form of a series
       dtype (str): data type for how to format the column based on the DataFormatter
       required (bool): if the column is required or not 
                        i.e. if the row should be delete if it is not filled
       pk (str): primary key label, indicating if the column is the primary required field

    """
    def __init__(self, dtype: str, required: bool, pk=False):
        self.series = None
        self.dtype = dtype
        self.required = required
        self.pk = pk

class DestinationTable(Logger):
    """
    Common object that all CDM objects (tables) inherit from.
    """

    @classmethod
    def from_df(cls,df,name=None):
        obj = cls(name)
        obj.__df = df
        #for colname in df.columns:
        #    obj[colname].series = df[colname]
        return obj

    def __len__(self):
        return len(self.__df)

    def reset(self):
        self.clear()
        self._meta.clear()
        self.__init_meta()
    
    def clear(self):
        self.__df = None
        #for field in self.fields:
        #    series = getattr(self,field)
        #    del series

    def __init_meta(self):
        self._meta = {'required_fields':{}}
        
    def __init__(self,name,_type,_version='v5_3_1',format_level=1):
        """
        Initialise the CDM DestinationTable Object class
        Args:
           _type (str): the name of the object being initialsed, e.g. "person"
           _version (str): the CDM version, see https://github.com/OHDSI/CommonDataModel/tags
        Returns: 
           None
        """
        self.name = name
        self._type = _type
        self.__init_meta()

        self.dtypes = DataFormatter()
        self.format_level = FormatterLevel(format_level)
        self.fields = self.get_field_names()
        #self.do_formatting = not format_level is None

        if len(self.fields) == 0:
            raise Exception("something misconfigured - cannot find any DataTypes for {self.name}")

        #print a check to see what cdm objects have been initialised
        self.logger.debug(self.get_destination_fields())
        self.__df = None

        #get the required fields
        self.required_fields = [
            field
            for field in self.get_field_names()
            if getattr(self,field).required == True
        ]
        
        self.automatically_fill_missing_columns = True
        self.tools = OperationTools()


    def get_field_names(self):
        """
        From the current object, loop over all member objects and find those that are instances
        of a DestinationField (column)
        
        Returns:
           list : a list of destination fields (columns [series])

        """
        return [
            item
            for item in self.__dict__.keys()
            if isinstance(getattr(self,item),DestinationField)
        ]
    
    def get_field_dtypes(self):
        """
        From the current object, loop over all member objects and find those that are instances
        of a DestinationField (column)
        
        Returns:
           list : a list of destination fields (columns [series])

        """
        return {
            item:getattr(self,item).dtype
            for item in self.__dict__.keys()
            if isinstance(getattr(self,item),DestinationField)
        }

    def get_ordering(self):
        """
        Loops over all associated fields and finds which have been marked as being a primary key.

        Returns:
            list: a string list of the names of primary columns (fields)
        """
        retval = [
            field
            for field in self.fields
            if getattr(self,field).pk == True
        ]

        return retval
        
    def __getitem__(self, key):
        """
        Retrieve a field (column) from the table (dataframe)

        Args:
           key (str) : name of a destination field
        Returns:
           DestinationField : the destination field object
        """
        
        return getattr(self, key)

    def __setitem__(self, key, obj):
        """
        Register a field object with the table
        """
        return setattr(self, key, obj)

    #def set_format_level(self,level):
    #    self.format_level = level
        
    def set_name(self,name):
        """
        Register/Set the name of the destination table
        """
        self.name = name
        self.logger.name = self.name
    
    def define(self,_):
        """
        define function, expected to be overloaded by the user defining the object
        """
        pass

    def get_destination_fields(self):
        """
        Get a list of all the destination fields that have been 
        loaded and associated to this cdm object

        
        Returns: 
           list: a list of all the destination fields that have been defined
        """
        return list(self.fields)

    def update(self,that):
        #extract all objects from the passed object
        objs = {k:v for k,v in that.__dict__.items() if k!='logger' }
        #add objects to this class
        self.__dict__.update(objs)
        

    def set_df(self,df):
        self.__df = df
    
    def get_df(self,force_rebuild=False,dont_build=False,dropna=False,**kwargs):
        """
        Retrieve a dataframe from the current object

        Returns:
           pandas.Dataframe: extracted dataframe of the cdm object
        """
       
        if not self.__df is None:
            self.logger.debug(f"df({hex(id(self.__df))}) already exists")
        
        if dont_build:
            if self.__df is None:
                self.__df = pd.DataFrame(columns = self.fields)
                self.set_df_name()
            return self.__df
        
        #if the dataframe has already been built.. just return it
        if not self.__df is None and not force_rebuild:
            self.logger.debug('already got a dataframe, so returning the existing one')
            if dropna:
                return self.__df.dropna(axis=1)
            else:
                return self.__df

        self.define(self)

        #get a dict of all series
        #each object is a pandas series
        dfs = {}

        for field in self.fields:
            obj = getattr(self,field)
            series = obj.series
            if series is None:
                #if required:
                #    self.logger.error(f"{field} is all null/none or has not been set/defined")
                #    raise RequiredFieldIsNone(f"{field} is a required for {self.name}.")
                continue

            #rename the column to be the final destination field name
            series = series.rename(field)
            #register the new series
            dfs[field] = series
            self.logger.debug(f'Adding series to dataframe from field "{field}"')

        #if there's none defined, dont do anything
        if len(dfs) == 0:
            self.logger.warning("no objects defined")
            self.__df = pd.DataFrame(columns = self.fields)
            self.set_df_name()
            return self.__df

        #check the lengths of the dataframes
        lengths = list(set([len(df) for df in dfs.values()]))
        if len(lengths)>1:
            self.logger.error("One or more inputs being mapped to this object has a different number of entries")
            for name,df in dfs.items():
                self.logger.error(f"{name} of length {len(df)}")
            raise BadInputs("Differring number of rows in the inputs")

        #create a dataframe from all the series objects
        df = pd.concat(dfs.values(),axis=1)

        #find which fields in the cdm havent been defined
        missing_fields = set(self.fields) - set(df.columns)

        #self._meta['defined_columns'] = df.columns.tolist()
        #self._meta['undefined_columns'] = list(missing_fields)
                
        #set these to a nan/null series
        for field in missing_fields:
            df[field] = np.NaN

        #simply order the columns 
        df = df[self.fields]

        df = self.finalise(df,**kwargs)
        df = self.format(df)
                    
        if dropna:
            df = df.dropna(axis=1)
         
        #register the df
        self.__df = df
        self.set_df_name()

        self.logger.info(f"created df ({hex(id(df))})[{self.get_df_name()}]")
        return self.__df


    def get_df_name(self):
        if not self.__df is None:
            return self.__df.attrs['name']
    
    def set_df_name(self):
        if self.__df is None:
            return
        name = re.sub("[^0-9a-zA-Z]+","_",self.name)
        self.__df.attrs['name'] = name
    
    def format(self,df):
        
        if self.format_level is FormatterLevel.OFF:
            self.logger.info('Not formatting data columns')
            return df
        elif self.format_level is FormatterLevel.ON:
            self.logger.info("Automatically formatting data columns.")
        elif self.format_level is FormatterLevel.CHECK:
            self.logger.info("Performing checks on data formatting.")


        for col in self.fields:
            #if is already all na/nan, dont bother trying to format
            is_nan_already = df[col].head(100).isna().all()
            if is_nan_already:
                continue

            obj = getattr(self,col)

            #dont try any formatting for primary keys that need to be integers
            if obj.pk == True or col == 'person_id':
                continue
            
            dtype = obj.dtype
            formatter_function = self.dtypes[dtype]
            
            nbefore = len(df[col])
            if nbefore == 0:
                self.logger.warning(f"trying to format an empty column ({cols})")
                
            nsample = 5 if nbefore > 5 else nbefore
            sample = df[col].sample(nsample)

            if self.format_level is FormatterLevel.ON:
                self.logger.debug(f"Formatting {col}")
                try:
                    df[col] = formatter_function(df[col])
                except Exception as e:
                    self.logger.critical(e)
                    if 'source_files' in self._meta:
                        self.logger.error("This is coming from the source file (table & column) ...")
                        self.logger.error(self._meta['source_files'][col])
                    raise(e)

                if col in self.required_fields:
                    df = df[~df[col].isna()]
                    #count the number of rows after
                    nafter = len(df)
                    self._meta['required_fields'][col]['after_formatting'] = nafter

                    if nafter == 0 :
                        self.logger.error(f"Something wrong with the formatting of the required field {col} using {dtype}")
                        self.logger.info(f"Formatting resulted in all NaN values. Sample of this column before formatting:")
                        self.logger.error(sample)
                        if 'source_files' in self._meta:
                            self.logger.error("This is coming from the source file (table & column) ...")
                            self.logger.error(self._meta['source_files'][col])
                        raise FormattingError(f"When formatting the required column {col}, using the formatter function {dtype}, all produced values are  NaN/null values.")
                    else:
                        ndiff = nafter - nbefore
                        if ndiff > 0:
                            self.logger.warning(f"Formatting of values in {col} removed {ndiff} rows, leaving {nafter} rows.")
                
            elif self.format_level is FormatterLevel.CHECK:
                self.logger.debug(f"Checking formatting of {col} to {dtype}")
                try:
                    _ = self.dtypes.check_formatting(df[col],formatter_function)
                except Exception as e:
                    if 'source_files' in self._meta:
                        self.logger.error("This is coming from the source file (table & column) ...")
                        self.logger.error(self._meta['source_files'][col])
                    raise(e) 

        return df

    def finalise(self,df,start_index=1,**kwargs):
        """
        Finalise a dataframe by dropping null/nan rows if a required field is missing.
        also sort the dataframe by the primary key of the table.

        Args:
            df (pandas.Dataframe): input dataframe
        Returns:
            pandas.Dataframe: cleaned output dataframe
        """
        
        #loop over the non-index fields
        for field in df.columns[1:]:
            #if it's not required, skip
            if field not in self.required_fields:
                continue
            
            #count the number of rows before
            nbefore = len(df)
            #remove rows which do not have this required field filled
            df = df[~df[field].isna()]
            
            #count the number of rows after
            nafter = len(df)
            #get the number of rows removed
            ndiff = nbefore - nafter
            #if rows have been removed
            if ndiff>0:
                #log a warning message if after requiring non-NaN values has removed all rows
                log = self.logger.warning if nafter > 0 else self.logger.error
                log(f"Requiring non-null values in {field} removed {ndiff} rows, leaving {nafter} rows.")

            #log some metadata
            self._meta['required_fields'][field] = {
                'before':nbefore,
                'after':nafter
            }

        #now index properly
        primary_column = df.columns[0]
        if primary_column != 'person_id':
            if df[primary_column].head(100).isnull().all():
                df[primary_column] = df.reset_index().index + start_index
            
        #return the dataframe sorted by the primary key requested
        #ordering = self.get_ordering()
        #if len(ordering) > 0:
        #    df = df.sort_values(self.get_ordering())
        return df




import pandas as pd
import datetime

class OperationTools:

    get_datetime = lambda self,df : pd.to_datetime(df,errors='coerce').dt.strftime('%Y-%m-%d %H:%M:%S.%f')
    get_date = lambda self,df : pd.to_datetime(df,errors='coerce').dt.strftime('%Y-%m-%d')
    get_year = lambda self,df : pd.to_datetime(df,errors='coerce').dt.year
    get_month = lambda self,df : pd.to_datetime(df,errors='coerce').dt.month
    get_day = lambda self,df : pd.to_datetime(df,errors='coerce').dt.day

    def keys(self):
        return [ key for key in dir(self) if key.startswith('get') ]
    
    def get_datetime_from_age(self,df,norm=None):
        if norm is None:
            #set normalisation to middle of 2020
            norm = datetime.datetime(2020, 7, 1)

        def _get_datetime_from_age(x):
            try:
                return norm - datetime.timedelta(days=365*int(float(x)))
            except Exception as e:
                print (e)
                return None
            
        retval = df.fillna(0).apply(lambda x: _get_datetime_from_age(x))
        return self.get_datetime(retval)

    def get_source_field_name_as_value(self,series):
        temp = series.copy()
        temp.values[:] = series.name
        return temp
    
    def make_scalar(self,series,value):
        #if we dont do a copy, the original data gets replaced
        temp = series.copy()
        temp.values[:] = value
        return temp

    def __getitem__(self, item):
        try:
            retval = getattr(self,item)
        except AttributeError as err:
            raise UnknownOperation(err)
        return retval



from graphviz import Digraph
import json

colorscheme = 'gnbu9'

def make_dag_from_objects(object_dict):
    _format = 'pdf'
    dot = Digraph(strict=True,format=_format)
    dot.attr(size='5,5')

    for destination_table,object_list in object_dict.items():
        dot.node(destination_table,shape='box')
        for obj in object_list:
            dot.node(obj,style='filled', fillcolor='yellow',shape='box')
            dot.edge(destination_table,obj,dir='forward')
                        
    dot.render('dag.gv', view=True)  
    return

def make_dag(data,output_file='dag.gv',format='pdf',render=False,orientation='RL',show_concepts=False):
    _format = format
    if render == False:
        _format = 'svg'

    dot = Digraph(strict=True,format=_format)
    dot.attr(rankdir=orientation)#, size='20,8')
            
    # #dot.attr(rankdir='RL', size='8,16',compound='true')
    
    with dot.subgraph(name='cluster_0') as dest, \
         dot.subgraph(name='cluster_1') as inp,\
         dot.subgraph(name='cluster_01') as c:
        
        #dest.attr(style='dotted',penwidth='4', label='CDM')
        #inp.attr(style='filled', fillcolor='lightgrey', penwidth='0', label='Input')
        
        dest.attr(style='filled', fillcolor='2', colorscheme='blues9', penwidth='0', label='Common Data Model')
        inp.attr(style='filled', fillcolor='2', colorscheme='greens9', penwidth='0', label='Source')
        if show_concepts:
            c.attr(style='filled', fillcolor='1', colorscheme='greens9', penwidth='0', label='Concepts')
        
        for destination_table_name,destination_tables in data.items():
            dest.node(destination_table_name,
                      shape='folder',style='filled',
                      fontcolor='white',colorscheme=colorscheme,
                      fillcolor='9')

            for ref_name,destination_table in destination_tables.items():
                for destination_field,source in destination_table.items():
                    source_field = source['source_field']
                    source_table = source['source_table']
                    
                    table_name = f"{destination_table_name}_{destination_field}"
                    dest.node(table_name,
                             label=destination_field,style='filled,rounded', colorscheme=colorscheme,
                             fillcolor='7',shape='box',fontcolor='white')

                    dest.edge(destination_table_name,table_name,arrowhead='none')

                    source_field_name =  f"{source_table}_{source_field}"
                    inp.node(source_field_name,source_field,
                             colorscheme=colorscheme,
                             style='filled,rounded',fillcolor='5',shape='box')
                    
                    if 'operations' in source:
                        operations = source['operations']

                    if 'term_mapping' in source and source['term_mapping'] is not None and show_concepts:
                        term_mapping = source['term_mapping']
                        if isinstance(term_mapping,dict):
                            concepts = list(term_mapping.values())
                        else:
                            concepts = [term_mapping]

                        for concept in concepts:
                            ref_name = ref_name.rsplit(" ",1)[0]
                            c.node(str(concept),label=ref_name,style='filled', colorscheme=colorscheme,
                                   fillcolor='black',shape='box',fontcolor='white')
                            dot.edge(table_name,str(concept),dir='back',color='red',penwidth='2')
                            dot.edge(str(concept),source_field_name,dir='back',color='red',penwidth='2')
                    else:                                                    
                        dot.edge(table_name,source_field_name,dir='back',penwidth='2')
                    
                    inp.node(source_table,shape='tab',fillcolor='4',colorscheme=colorscheme,style='filled')
                    inp.edge(source_field_name,source_table,arrowhead='none')#,dir='back',arrowhead='none')


    #dot.subgraph(destinations)
    #dot.subgraph(sources)

    #destinations = dot.subgraph(name='cdm')
    #destinations = Digraph('cdm')
    #destinations.attr(style='dotted',rank='same',label='process #2')
    #dot.subgraph(destinations)
    #sources = Digraph('sources')
    #sources.attr(style='dotted',rank='same',label='process #1')

    
                
    if render:
        dot.render(output_file, view=True)  
        return
    #    
    return dot.pipe().decode('utf-8')

def make_report_dag(data,name='dag',render=False,orientation='RL'):
    _format = 'svg'
    if render == True:
        _format = 'pdf'

    dot = Digraph(strict=True,format=_format)
    dot.attr(rankdir=orientation)#, size='20,8')

    colorscheme = 'bupu9'
    colorscheme = 'rdbu10'
    
    for table in data:

        dataset=table['meta']['dataset']
        dot.node(dataset,
             style='filled', colorscheme=colorscheme,
             fillcolor='10',shape='box',fontcolor='white')
        
        table_name = table['table']

        dot.node(table_name,
                  style='filled', colorscheme=colorscheme,
                  fillcolor='7',shape='box')#,fontcolor='white')
        n = table['meta']['nscanned']
        dot.edge(table_name,dataset,arrowhead='none',label=f"N={n}")
        
        for field in table['fields']:
            field_name = field['field']

            ref_name = table_name+field_name
            dot.node(ref_name,label=field_name,
                  style='filled,rounded', colorscheme=colorscheme,
                  fillcolor='6',shape='box')#,fontcolor='white')
            dot.edge(ref_name,table_name,arrowhead='none')
    
            for value in field['values']:
                frequency = int(round(value['frequency']*100.,0))
                if frequency < 5:
                    continue
                frequency = f"{frequency} %"

                value_name = value['value']
                ref_name_v = table_name+field_name+value_name
                dot.node(ref_name_v,label=value_name,
                         style='filled,rounded', colorscheme=colorscheme,
                         fillcolor='5',shape='box')#,fontcolor='white')
                dot.edge(ref_name_v,ref_name,label=frequency,arrowhead='none')

                if not 'concepts' in value:
                    continue

                for concept in value['concepts']:
                    concept_id = concept['concept_id']
                    concept_name = concept['concept_name']
                    domain_id = concept['domain_id']
                    label = concept_name+"\n"+str(concept_id)+' | '+domain_id
                    
                    ref_name_v2 = str(concept_id)+ ref_name_v
                    dot.node(ref_name_v2,label=label,
                             style='filled,rounded', colorscheme=colorscheme,
                             fillcolor='3',shape='box',fontcolor='white')
                    dot.edge(ref_name_v2,ref_name_v,arrowhead='none')

                    #dot.node(domain_id,
                    #         style='filled,rounded', colorscheme=colorscheme,
                    #         fillcolor='2',shape='box',fontcolor='black')
                    #dot.edge(domain_id,ref_name_v2,arrowhead='none')

    
    if render:
        dot.render(f'{name}.gv', view=True)  
        return




class TableNotFoundError(Exception):
    pass
class FieldNotFoundError(Exception):
    pass

def get_person_ids(rules):
    return {
        subtable['person_id']['source_table']:subtable['person_id']['source_field'] 
        for table in rules['cdm'].values() 
        for subtable in table.values()
    }


def get_source_field(table,name):
    if name not in table:
        if name.lower() in table:
            return table[name.lower()]
        else:
            raise FieldNotFoundError(f"Cannot find {name} in table {table.name}. Options are {table.columns.tolist()}")
    return table[name]

def get_source_table(inputs,name):
    #make a copy of the input data column slice
    if name not in inputs.keys():
        short_keys = {key[:31]:key for key in inputs.keys()}
        if name in short_keys:
            name = short_keys[name]
        elif name.lower() in short_keys:
            name = short_keys[name.lower()]
        else:
            raise TableNotFoundError(f"Cannot find {name} in inputs. Options are {inputs.keys()}")
    table = inputs[name]
    table.name = name
    return inputs[name]


def load_from_file(this):
    df = this.inputs[this.fname].dropna(axis=1)
    for colname in df.columns:
        this[colname].series = df[colname]


def apply_rules(this,rules,inputs=None):
    this.logger.info("Called apply_rules")

    if inputs is None:
        inputs = this.inputs

    this._meta['source_files'] = {}
    for destination_field,rule in rules.items():
        source_table_name = rule['source_table']
        source_field_name = rule['source_field']
        operations = None
        if 'operations' in rule:
            operations = rule['operations']
        term_mapping = None
        if 'term_mapping' in rule:
            term_mapping = rule['term_mapping']

        source_table = get_source_table(inputs,source_table_name)
        source_field = get_source_field(source_table,source_field_name)
        series = source_field.copy()

        if operations is not None:
            for operation in operations:
                function = this.tools[operation]
                series = function(series)
                
        if term_mapping is not None:
            if isinstance(term_mapping,dict):
                # value level mapping
                # - term_mapping is a dictionary between values and concepts
                # - map values in the input data, based on this map

                #need to make the value a string for mapping
                #pandas has a weird behaviour that when the value is an Int
                #the resulting series is a float64
                term_mapping = {k:str(v) for k,v in term_mapping.items()}
                series = series.map(term_mapping)
            else:
                # field level mapping.
                # - term_mapping is the concept_id
                # - set all values in this column to it
                series.values[:] = term_mapping

        this[destination_field].series = series
        this._meta['source_files'][destination_field] = {'table':source_table_name,'field':source_field_name}
        this.logger.info(f"Mapped {destination_field}")



import os
import glob
import copy
import json
import pandas as pd
from carrot.tools.logger import _Logger as Logger
import carrot.io as io

class MissingInputFiles(Exception):
    pass
class DifferingColumns(Exception):
    pass
class DifferingRows(Exception):
    pass

                        
    
def load_json_delta(f_in,original):
    logger = Logger("load_json_delta")
    
    data = load_json(f_in)

    if isinstance(original,str):
        original = load_json(original)
    
    if original == None:
        return data

    if original == data:
        return data

    logger.info(f"loading a json from '{f_in}' as a delta")

    
    original_date = original['metadata']['date_created']
    data_date = data['metadata']['date_created']

    logger.info(f"Original JSON date: {original_date}")
    logger.info(f"New JSON date: {data_date}")

    _data = copy.deepcopy(data)

    for destination_table,rule_set in data['cdm'].items():
        for name,rules in rule_set.items():
            exists_in_original_rules = None
            if destination_table in original['cdm']:
                if name in original['cdm'][destination_table]:
                    exists_in_original_rules = original['cdm'][destination_table][name]
            
            if exists_in_original_rules:
                _data['cdm'][destination_table].pop(name)
            else:
                logger.info(f"Detected a new rule for {name}")
                logger.debug(json.dumps(rules,indent=6))
        
        if not _data['cdm'][destination_table]:
            _data['cdm'].pop(destination_table)

    logger.debug(json.dumps(_data,indent=6))
    return _data
    

def load_json(f_in):

    if os.path.exists(f_in):
        data = json.load(open(f_in))
    else:
        try:
            data = json.loads(f_in)
        except Exception as err:
            raise FileNotFoundError(f"{f_in} not found. Or cannot parse as json")

    return data


def create_csv_store(**kwargs):
    return io.LocalDataCollection(**kwargs)

def create_bclink_store(**kwargs):
    return io.BCLinkDataCollection(**kwargs)

def create_sql_store(**kwargs):
    return io.SqlDataCollection(**kwargs)    

def load_sql(**kwargs):
    store = create_sql_store(**kwargs)
    return store

def load_csv(_map,chunksize=None,
             dtype=str,nrows=None,
             lower_col_names=False,
             load_path="",
             rules=None,
             sep=',',
             na_values=['']):

    if isinstance(_map,list):
        _map = {
            os.path.basename(x):x
            for x in _map
        }
    
    logger = Logger("carrot.tools.load_csv")

    if isinstance(_map,list) or isinstance(_map,tuple):
        _map = { x:x for x in _map}
    elif isinstance(_map,str):
        _map = { _map:_map }
        
    if rules is not None:
        logger.debug("rules .json file supplied")
        if not isinstance(rules,dict):
            rules = load_json(rules)

        inputs_from_cli = list(_map.keys())
            
        source_map = get_mapped_fields_from_rules(rules)
        inputs_from_json = list(source_map.keys())

        if len(inputs_from_cli) == 0:
            raise MissingInputFiles (f"You haven't loaded any input files!")
            
        logger.debug(f"{len(inputs_from_cli)} input files loaded")
        logger.debug(f"{inputs_from_cli}")
        
        missing_inputs = list(set(inputs_from_json) - set(inputs_from_cli))
        if len(missing_inputs) > 0 :
            raise MissingInputFiles (f"Found the following files {missing_inputs} in the json file, that are not in the loaded file list... {inputs_from_cli}")
        
        #reduce the mapping of inputs, if we dont need them all
        _map = {
            k: {
                'file':v,
                'fields':source_map[k]
            }
            for k,v in _map.items()
            if k in source_map
        }


    if not nrows is None:
        chunksize = nrows if chunksize is None else chunksize

    retval = io.LocalDataCollection(chunksize=chunksize)

    for key,obj in _map.items():
        fields = None
        if isinstance(obj,str):
            fname = obj
        else:
            fname = obj['file']
            fields = obj['fields']

        df = pd.read_csv(load_path+fname,
                         chunksize=chunksize,
                         #iterator=True,
                         nrows=nrows,
                         sep=sep,
                         keep_default_na=False,
                         na_values=na_values,
                         dtype=dtype,
                         usecols=fields)

        df.attrs = {'original_file':load_path+fname}

        
        if isinstance(df,pd.DataFrame):
            #this should be removed
            if lower_col_names:
                df.columns = df.columns.str.lower()

        retval[key] = io.DataBrick(df,name=key)

    return retval


def load_tsv(_map,**kwargs):
    kwargs['sep']="\t"
    return load_csv(_map,**kwargs)


def get_subfolders(input_folder):
    return { 
        os.path.basename(f.path):f.path 
        for f in os.scandir(input_folder) 
        if f.is_dir() and not os.path.basename(f.path).startswith('.')
    }

def get_files(path,type='csv'):
    return [x.path for x in os.scandir(path) if x.path.endswith(f'.{type}')]

def get_file_map_from_dir(_dir,ext='.csv'):
    if not os.path.isdir(_dir):
        _dir = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),'..','data',_dir)
        )

    _map = {}
    for fname in glob.glob(f"{_dir}{os.path.sep}*{ext}"):
        key = os.path.basename(fname)
        _map[key] = fname
    
    return _map
 
def remove_missing_sources_from_rules(rules,tables):
    logger = Logger("remove_missing_sources_from_rules")

    tables = [os.path.basename(x) for x in tables]
   
    rules_copy = copy.deepcopy(rules)

    for destination_table,cdm_table in rules['cdm'].items():
        for table_name,sub_table in cdm_table.items():
            first = list(sub_table.keys())[0]
            source_table = sub_table[first]['source_table']
            if source_table not in tables:
                rules_copy['cdm'][destination_table].pop(table_name)
                logger.warning(f"removed {table_name} from rules because it was not loaded")
                
        if not rules_copy['cdm'][destination_table]:
            rules_copy['cdm'].pop(destination_table)
            logger.warning(f"removed cdm table '{destination_table}' from rules")
        
    return rules_copy

def filter_rules_by_destination_tables(rules,tables):
    rules_copy = copy.deepcopy(rules)

    for destination_table,cdm_table in rules['cdm'].items():
        if not destination_table in tables:
            rules_copy['cdm'].pop(destination_table)
           
    return rules_copy

def filter_rules_by_object_names(rules,names):
    rules_copy = copy.deepcopy(rules)
    for destination_table,cdm_table in rules['cdm'].items():
        for object_name,rules in cdm_table.items():
            if not object_name in names:
                rules_copy['cdm'][destination_table].pop(object_name)
        if not rules_copy['cdm'][destination_table]:
            rules_copy['cdm'].pop(destination_table)
           
    return rules_copy
    
def get_source_tables_from_rules(rules,table):
    sources = [ 
        x['source_table']
        for dest_table,cdm_obj_set in rules['cdm'].items()
        for cdm_obj in cdm_obj_set.values()
        for x in cdm_obj.values()
        if dest_table == table
    ]
    return list(set(sources))

def get_mapped_fields_from_rules(rules):
    #extract a tuple of source tables and source fields
    sources = [
        (x['source_table'],x['source_field'])
        for cdm_obj_set in rules['cdm'].values()
        for cdm_obj in cdm_obj_set.values()
        for x in cdm_obj.values()
    ]
    
    source_map = {}
    for (table,field) in sources:
        if table not in source_map:
            source_map[table] = []
        source_map[table].append(field)
            
    source_map = {
        k:list(set(v))
        for k,v in source_map.items()
    }

    return source_map
    

def get_separator_from_filename(fname):
    _, fileExtension = os.path.splitext(fname)
    if fileExtension == '.tsv':
        return '\t'
    else:
        return ','


def diff_csv(file1,file2,separator=None,nrows=None):
    logger = Logger("CSV File Diff")
    
    if separator == None:
        sep1 = get_separator_from_filename(file1)
        sep2 = get_separator_from_filename(file2)
    else:
        sep1 = separator
        sep2 = separator

    df1 = pd.read_csv(file1,sep=sep1,nrows=nrows)
    df2 = pd.read_csv(file2,sep=sep2,nrows=nrows)
    
    exact_match = df1.equals(df2)
    if exact_match:
        return

    df = pd.concat([df1,df2]).drop_duplicates(keep=False)

    if len(df) > 0:
        logger.error(" ======== Differing Rows ========== ")
        logger.error(df)
        m = df1.merge(df2, on=df.columns[0], how='outer', suffixes=['', '_'], indicator=True)[['_merge']]
        m = m[~m['_merge'].str.contains('both')]
        file1 = file1.split('/')[-1]
        file2 = file2.split('/')[-1]
        
        m['_merge'] = m['_merge'].map({'left_only':file1,'right_only':file2})
        m = m.rename(columns={'_merge':'Only Contained Within'})
        m.index.name = 'Row Number'
        logger.error(m.reset_index().to_dict(orient='records'))
        raise DifferingRows("Something not right with the rows, changes detected.")
        
    elif len(df1.columns) != len(df2.columns):
        
        raise DifferingColumns('in df1 but not df2',list(set(df1.columns) - set(df2.columns)),'\n',
                               'in df2 but not df1',list(set(df2.columns) - set(df1.columns)))

    else:
        logger.error(" ======= Rows are likely in a different order ====== ")
        for i in range(len(df1)):
            if not (df1.iloc[i] == df2.iloc[i]).any():
                print ('Row',i,'is in a different location')
        raise Exception("differences detected")

    



import pandas as pd
from carrot.tools.logger import Logger
from types import GeneratorType
import io

class DataCollection(Logger):
    def __init__(self,chunksize=None,nrows=None,**kwargs):
        self.logger.info("DataCollection Object Created")
        self.__bricks = {}
        self.chunksize = chunksize
        self.nrows = nrows
        
        if self.chunksize is not None:
            self.logger.info(f"Using a chunksize of '{self.chunksize}' nrows")

    def print(self):
        print (self.all())
        
    def all(self):
        return {
            key:self[key]
            for key in self.keys()
        }
    def finalise(self):
        pass
        
    def keys(self):
        return self.__bricks.keys()

    def items(self):
        return self.__bricks.items()

    def __setitem__(self,key,obj):
        self.logger.info(f"Registering  {key} [{obj}]")
        self.__bricks[key] = obj

    def load_global_ids(self):
        return

    def load_indexing(self):
        return
            
    def next(self):
        #loop over all loaded files
        self.logger.info("Getting next chunk of data")

        used_bricks = []
        for key,brick in self.items():
            if brick.is_finished():
                continue
            
            if brick.is_init():
                used_bricks.append(brick)
            else:
                continue
                
            self.logger.info(f"Getting the next chunk of size '{self.chunksize}' for '{key}'")
            brick.get_chunk(self.chunksize)
            n = len(brick.get_df())
            self.logger.info(f"--> Got {n} rows")
            if n == 0:
                brick.set_finished(True)
            
        #check if all __dataframe objects are empty
        #if they are, raise a StopIteration as processing has finished
        if all([x.is_finished() for x in used_bricks]):
            self.logger.info("All input files for this object have now been used.")
            raise StopIteration
        
                           
    def get_handler(self,key):
        brick = self.__bricks[key]
        return brick.get_handler()

    def get_all(self):
        self.logger.info(f"Retrieving initial dataframes for the first time")
        for b in self.__bricks.values():
            b.get_chunk(self.chunksize)
            b.set_init(True)

    def get(self,key):
        return self.__bricks[key]
            
    def __getitem__(self,key):
        brick = self.__bricks[key]
        if not brick.is_init():
            self.logger.info(f"Retrieving initial dataframe for '{key}' for the first time")
            brick.get_chunk(self.chunksize)
            brick.set_init(True)
            
        #if any(not x.is_init() for x in self.__bricks.values()):
        #    self.get_all()
                
        df = brick.get_df()
        self.logger.debug(f"Got brick {brick}")
        return df
    
    def reset(self):
        self.logger.info(f"resetting used bricks")
        for key,brick in self.items():
            brick.reset()

class DataBrick:
    def __init__(self,df_handler,name=None):
        self.name = name
        self.__df_handler = df_handler
        self.__df = None
        self.__end = False
        self.__is_init = False

    def get_handler(self):
        return self.__df_handler

    def is_finished(self):
        return self.__end
    
    def set_finished(self,value):
        self.__end = value

    def is_init(self):
        return self.__is_init

    def set_init(self,value):
        self.__is_init = value
    
    def reset(self):
        if isinstance(self.__df_handler,pd.io.parsers.TextFileReader):
            options = self.__df_handler.orig_options
            f = self.__df_handler.f
            del self.__df_handler
            options['engine'] = 'c'
            if isinstance(f,io.StringIO):
                f.seek(0)
            self.__df_handler = pd.io.parsers.TextFileReader(f,**options)
            
        self.__df = None
        self.__end = False
        self.__is_init = False
        return True
    
    def get_chunk(self,chunksize):
        if self.__end == True:
            return
        #if the df handler is a TextFileReader, get a dataframe chunk
        if isinstance(self.__df_handler,pd.io.parsers.TextFileReader):
            try:
                #for this file reader, get the next chunk of data
                self.__df = self.__df_handler.get_chunk(chunksize)
            except StopIteration:#,ValueError):
                #otherwise, if at the end of the file reader, return an empty frame
                self.__df = pd.DataFrame(columns=self.__df.columns) if self.__df is not None else None
                self.__end = True
        elif isinstance(self.__df_handler,pd.DataFrame):
            #if we're handling non-chunked data
            if self.__df is not None:
                #return an empty dataframe if we've already loaded this dataframe
                self.__df = pd.DataFrame(columns=self.__df.columns)
            else:
                #otherwise return the dataframe as it's the first time we're getting it
                self.__df = self.__df_handler
            self.__end = True
        elif isinstance(self.__df_handler, GeneratorType):
            try:
                self.__df = next(self.__df_handler)
            except StopIteration:
                self.__df = pd.DataFrame(columns=self.__df.columns) if self.__df is not None else None
                self.__end = True
        else:
            raise NotImplementedError(f"{type(self.__df_handler)} not implemented")

    def get_df(self):
        return self.__df



from .model import CommonDataModel
from .operations import OperationTools

from .objects import get_cdm_class, get_cdm_decorator

#can we make these imports dynamic?
from .objects import (
    Person,
    ConditionOccurrence,
    VisitOccurrence,
    Measurement,
    Observation,
    DrugExposure
)

from .decorators import (
    define_table,
    define_person,
    define_condition_occurrence,
    define_visit_occurrence,
    define_measurement,
    define_observation,
    define_drug_exposure,
    load_file,
    from_table,
    qsub
)



import carrot 
import importlib
import inspect

__default_cdm_version = '5.3.1'
if 'cdm' in carrot.params:
    __cdm_version = carrot.params['cdm']
else:
    __cdm_version = __default_cdm_version

__cdm_version_split = '_'.join(__cdm_version.split('.'))

try:
    __cdm_tables = importlib.import_module(f'carrot.cdm.objects.versions.v{__cdm_version_split}')
except ModuleNotFoundError as e:
    raise ModuleNotFoundError(f'Cannot find CDM version {__cdm_version}, this does not exist in this package!') from e

__cdm_tables = {
    m[0]:getattr(__cdm_tables,m[0])
    for m in inspect.getmembers(
            __cdm_tables,
            inspect.isclass)
}


__cdm_object_map = {
    obj.name : obj
    for obj in __cdm_tables.values()
}
def get_cdm_tables():
    return __cdm_object_map

import sys
this = sys.modules[__name__]

for name,_class in __cdm_tables.items():
    setattr(this,name,_class)

from .. import decorators

__cdm_decorator_map = {
    name.replace("define_",""):getattr(decorators,name)
    for name in dir(decorators)
    if 'define' in name
}

def get_cdm_decorator(key):
    return __cdm_decorator_map[key]

def get_cdm_class(key):
    return __cdm_object_map[key]

from .common import DestinationTable, DataFormatter, FormatterLevel




import pandas as pd
from ...common import DestinationTable, DestinationField

class ConditionOccurrence(DestinationTable):
    """
    CDM Condition Occurrence object class
    """
    
    name = 'condition_occurrence'
    def __init__(self,name=None):
        self.condition_occurrence_id       = DestinationField(dtype="Integer"   , required=True , pk=True)
        self.person_id                     = DestinationField(dtype="Integer"   , required=True )
        self.condition_concept_id          = DestinationField(dtype="Integer"   , required=True )
        self.condition_start_date          = DestinationField(dtype="Date"      , required=False )
        self.condition_start_datetime      = DestinationField(dtype="Timestamp" , required=True )
        self.condition_end_date            = DestinationField(dtype="Date"      , required=False )
        self.condition_end_datetime        = DestinationField(dtype="Timestamp" , required=False )
        self.condition_type_concept_id     = DestinationField(dtype="Integer"   , required=False )
        self.stop_reason                   = DestinationField(dtype="Text20"    , required=False )
        self.provider_id                   = DestinationField(dtype="Integer"   , required=False )
        self.visit_occurrence_id           = DestinationField(dtype="Integer"   , required=False )
        self.condition_source_value        = DestinationField(dtype="Text50"    , required=False )
        self.condition_source_concept_id   = DestinationField(dtype="Integer"   , required=False )
        self.condition_status_source_value = DestinationField(dtype="Text50"    , required=False )
        self.condition_status_concept_id   = DestinationField(dtype="Integer"   , required=False )
        
        if name is None:
            name = hex(id(self))
        super().__init__(name,self.name)

        
    def get_df(self,**kwargs):
        """
        Overload/append the creation of the dataframe, specifically for the condition_occurrence objects
        * condition_concept_id is required to be not null
          this can happen when spawning multiple rows from a person
          we just want to keep the ones that have actually been filled
        
        Returns:
           pandas.Dataframe: output dataframe
        """

        df = super().get_df(**kwargs)
        if self.automatically_fill_missing_columns == True:
            if df['condition_start_date'].isnull().all():
                df['condition_start_date'] = self.tools.get_date(df['condition_start_datetime'])

            if df['condition_end_date'].isnull().all():
                df['condition_end_date'] = self.tools.get_date(df['condition_end_datetime'])
        return df
    



import setuptools
import os
import sys

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
    print (long_description)
cwd = os.getcwd()
os.system(f'ls {cwd}')
with open('requirements.txt') as f:
    required = f.read().splitlines()


sys.path.append("carrot/")
from _version import __version__ as version

    
setuptools.setup(
    name="carrot-cdm", 
    author="CO-CONNECT Collaboration",
    version=version,
    author_email="calmacx@gmail.com",
    description="Python package for performing mapping of ETL to CDM ",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/HDRUK/CaRROT-CDM",
    entry_points = {
        'console_scripts':[
            'carrot=carrot.cli.cli:carrot'
        ],
    },
    packages=setuptools.find_packages(),
    extras_require = {
        'airflow':['apache-airflow'],
        'performance':['snakeviz'],
    },
    install_requires=required,
    package_data={'carrot': ['data/cdm/*','data/example/*/*','data/test/*/*','data/test/*/*/*']},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import os
import pandas as pd
import numpy as np
import json
import copy
import getpass
import operator

import shutil
import threading
import concurrent.futures
from time import gmtime, strftime, sleep, time

from .operations import OperationTools
from carrot.tools.logger import Logger
from carrot.tools.profiling import Profiler
import carrot.tools 
from carrot.io import DataCollection

from carrot import __version__ as carrot_version
from .objects import DestinationTable, FormatterLevel
from .objects import get_cdm_class, get_cdm_decorator
from .decorators import load_file, analysis

class BadInputObject(Exception):
    pass

class PersonExists(Exception):
    pass


class CommonDataModel(Logger):
    """Pythonic Version of the OHDSI CDM.

    This class controls and manages CDM Table objects that are added to it

    When self.process() is executed by the user, all added objects are defined, merged, formatted and finalised, before being dumped to an output file (.tsv file by default).

    """

    
    @classmethod
    def load(cls,inputs,**kwargs):
        default_kwargs = {'save_files':False,'do_mask_person_id':False,'format_level':0}
        default_kwargs.update(kwargs)
        cdm = cls(**default_kwargs)
        cdm._load_inputs(inputs)
        return cdm

    @classmethod
    def from_rules(cls,rules,**kwargs):
        cdm = cls(**kwargs)
        cdm.create_and_add_objects(rules)
        return cdm
    

    def __init__(self, name=None, omop_version='5.3.1',
                 outputs = None,
                 save_files=True,
                 inputs=None,
                 use_profiler=False,
                 format_level=None,
                 do_mask_person_id=True,
                 drop_duplicates=True,
                 automatically_fill_missing_columns=True):
        """
        CommonDataModel class initialisation 
        Args:
            name (str): Give a name for the class to appear in the logging
            output_folder (str): Path of where the output tsv/csv files should be written to.
                                 The default is to save to a folder in the current directory
                                 called 'output_data'.
            inputs (dict or DataCollection): inputs can be a dictionary mapping file names to pandas dataframes,
                                        or can be a DataCollection object
            use_profiler (bool): Turn on/off profiling of the CPU/Memory of running the current process. 
                                 The default is set to false.
        """
        self.profiler = None
        name = self.__class__.__name__ if name is None else self.__class__.__name__ + "::" + name
            
        self.logger.info(f"CommonDataModel ({omop_version}) created with co-connect-tools version {carrot_version}")

        self.omop_version = omop_version

        self.drop_duplicates = drop_duplicates
        self.do_mask_person_id = do_mask_person_id
        self.execution_order = None
        
        if format_level == None:
            format_level = 1
        try:
            format_level = int(format_level)
        except ValueError:
            self.logger.error(f"You as specifying format_level='{format_level}' -- this should be an integer!")
            raise ValueError("format_level not set as an int ")
        
        self.format_level = FormatterLevel(format_level)
        self.profiler = None

        self.outputs = outputs
        self.save_files = save_files

        if use_profiler:
            self.logger.debug(f"Turning on cpu/memory profiling")
            self.profiler = Profiler(name=name)
            self.profiler.start()

        #perform some checks on the input data
        if isinstance(inputs,dict):
            self.logger.info("Running with an DataCollection object")
        elif isinstance(inputs,DataCollection):
            self.logger.info("Running with an DataCollection object")
        elif inputs is not None:
            _type = type(inputs).__name__
            raise BadInputObject(f"input object {inputs} is of type {_type}, which is not a valid input object")

        if inputs is not None:
            if hasattr(self,'inputs'):
                self.logger.warning("overwriting inputs")
            self.inputs = inputs
        elif not hasattr(self,'inputs'):
            self.inputs = None
            
        #register opereation tools
        self.tools = OperationTools()

        #allow rules to be generated automatically or not
        self.automatically_fill_missing_columns = automatically_fill_missing_columns
        if self.automatically_fill_missing_columns:
            self.logger.info(f"Turning on automatic cdm column filling")

        #define a person_id masker, if the person_id are to be masked
        if self.outputs:
            self.person_id_masker = self.outputs.load_global_ids()
            self.indexing_conf = self.outputs.load_indexing()
        else:
            self.person_id_masker = None
            self.indexing_conf = None

        #stores the final pandas dataframe for each CDM object
        # {
        #   'person':pandas.DataFrame,
        #   'measurement':pandas.DataFrame.
        #   ....
        #}
        self.__df_map = {}

        #stores the invididual objects associated to this model
        # {
        #     'observation':
        #     {
        #         'observation_0': <carrot.cdm.objects.observation.Observation object 0x000>,
        #         'observation_1': <carrot.cdm.objects.observation.Observation object 0x001>,
        #     },
        #     'measurement':
        #     {
        #         'measurement_0': <carrot.cdm.objects.measurement.Measurement object 0x000>,
        #         ...
        #     }
        #     ...
        # }
        self.__objects = {}
        #check if objects have already been registered with this class
        #via the decorator methods
                
        registered_objects = [
            getattr(self,name)
            for name in dir(self)
            if isinstance(getattr(self,name),DestinationTable)
        ]
        #if they have, then include them in this model
        for obj in registered_objects:
            obj.inputs = self.inputs
            self.add(obj)

        self.__analyses = {
            name:getattr(self,name)
            for name in dir(self)
            if isinstance(getattr(self,name),analysis)
        }
                    
        #bookkeep some logs
        self.logs = {
            'meta':{
                'version': carrot_version,
                'created_by': getpass.getuser(),
                'created_at': strftime("%Y-%m-%dT%H%M%S", gmtime()),
                'dataset':name,
                'total_data_processed':{}
            }
        }
        
    def _load_inputs(self,inputs):
        for fname in inputs.keys():
            destination_table,_ = os.path.splitext(fname)
            try:
                obj = get_cdm_class(destination_table).from_df(inputs[fname],destination_table)
            except KeyError:
                self.logger.warning(f"Not loading {fname}, this is not a valid CDM Table")
                continue
                
            df = obj.get_df(force_rebuild=False)
            self[destination_table] = df.set_index(df.columns[0])

    def reset(self):
        self.__df_map.clear()
        [x.reset() for x in self.get_all_objects()]
        self.inputs.reset()

        if self.outputs:
            self.person_id_masker = self.outputs.load_global_ids()
            self.indexing_conf = self.outputs.load_indexing()
        else:
            self.person_id_masker = None
            self.indexing_conf = None

        
        
    def close(self):
        """
        Class destructor:
              Stops the profiler from running before deleting self
        """


        self.logger.info(json.dumps(self.logs['meta'],indent=6))
        if self.outputs:
            self.outputs.write_meta(self.logs)
            self.outputs.finalise()

        if not hasattr(self,'profiler'):
            return
        if self.profiler:
            self.profiler.stop()
            df_profile = self.profiler.get_df()
            f_out = self.output_folder
            f_out = f'{f_out}{os.path.sep}logs{os.path.sep}'
            if not os.path.exists(f'{f_out}'):
                self.logger.info(f'making output folder {f_out}')
                os.makedirs(f'{f_out}')
                
            date = self.logs['meta']['created_at']
            fname = f'{f_out}{os.path.sep}statistics_{date}.csv'
            df_profile.to_csv(fname)
            self.logger.info(f"Writen the memory/cpu statistics to {fname}")
            self.logger.info("Finished")

    @classmethod
    def from_existing(cls,**kwargs):
        """
        Initialise the CDM model from existing data in the CDM format
        """
        cdm = cls(**kwargs)
        if 'inputs' not in kwargs:
            raise NoInputFiles("you need to specify some inputs")
        inputs = kwargs['inputs']
        #loop over all input names
        for fname in inputs.keys():
            #obtain the name of the destination table
            #e.g fname='person.tsv' we want 'person'
            destination_table,_ = os.path.splitext(fname)
            name = destination_table
            if '.' in destination_table:
                destination_table = destination_table.split('.')[0]
            try:
                obj = get_cdm_class(destination_table).from_df(inputs[fname],name=name)
            except KeyError:
                cdm.logger.warning(f"Not loading {fname}, this is not a valid CDM Table")
                continue
            cdm.add(obj)
        return cdm

    def __del__(self):
        self.__df_map.clear()
        del self.__df_map
        self.__objects.clear()
        del self.__objects
                
        
    def __getitem__(self,key):
        """
        Ability lookup processed objects from the CDM
        Example:
            cdm = CommonDataModel()
            ...
            cdm.process()
            ...
            person = cdm['person']
        Args:
            key (str): The name of the cdm table to be returned
        Returns:
            pandas.DataFrame if a processed object is found, otherwise returns None
        """
        if key not in self.__df_map.keys():
            return None
        else:
            return self.__df_map[key]

    def __setitem__(self,key,obj):
        """
        Registration of a new dataframe for a new object
        Args:
            key (str) : name of the CDM table (e.g. "person")
            obj (pandas.DataFrame) : dataframe to refer to 
        """
        self.logger.debug(f"creating {obj} for {key}")
        self.__df_map[key] = obj

    def print(self):
        for name in self.keys():
            print (self[name].dropna(axis=1))
        
    def add(self,obj):
        """
        Function to add a new CDM table (object) to the current model
        Args:
            obj (DestinationTable) : CDM Table to be registered with the class
        """
        if obj._type not in self.__objects:
            self.__objects[obj._type] = {}
            
        if obj.name in self.__objects[obj._type].keys():
            raise Exception(f"Object called {obj.name} already exists")

        obj.cdm = self
        obj.format_level = self.format_level
        
        self.__objects[obj._type][obj.name] = obj
        self.logger.info(f"Added {obj.name} of type {obj._type}")


    def create_and_add_objects(self,config):
        #loop over the cdm object types defined in the configuration
        #e.g person, measurement etc..
        for destination_table,rules_set in config['cdm'].items():
            #loop over each object instance in the rule set
            #for example, condition_occurrence may have multiple rulesx
            #for multiple condition_ocurrences e.g. Headache, Fever ..
            for name,rules in rules_set.items():
                #make a new object for the cdm object
                #Example:
                # destination_table : person
                # get_cdm_class returns <Person>
                # obj : Person()
                obj = get_cdm_class(destination_table)()
                #set the name of the object
                obj.set_name(name)
            
                #Build a lambda function that will get executed during run time
                #and will be able to apply these rules to the inputs that are loaded
                #(this is useful when chunk)
                obj.define = lambda x,rules=rules : carrot.tools.apply_rules(x,rules,inputs=self.inputs)
                
                #register this object with the CDM model, so it can be processed
                self.add(obj)

        
    def add_analysis(self,func,_id=None):
        if _id is None:
            _id = hex(id(func))
        self.__analyses[_id] = func
       
    def get_analyses(self):
        return self.__analyses
    def get_analysis(self,key):
        return self.__analyses[key]
        
    def run_analysis(self,f):
        return f(self)
        
    def run_analyses(self,analyses=None,max_workers=4):
            
        def msg(x):
            self.logger.info(f"finished with {x}")
            self.logger.debug(x.result())

        start = time()
            
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = {}
            for name,f in self.__analyses.items():
                self.logger.info(f"Start thread for {f} ")
                future = executor.submit(f, self)
                future.add_done_callback(msg)
                futures[future] = name

            while True:
                status = {futures[f]:{'status':'running' if f.running() else 'done' if f.done() else 'waiting'} for f in futures}
                self.logger.debug(json.dumps(status,indent=6))
                if all([f.done() for f in futures]):
                    break
                sleep(1)
                
                
        results = {}
        for future in concurrent.futures.as_completed(futures):
            _id = futures[future]
            results[_id] = future.result()

        end = time() - start
        self.logger.info(f"Running analyses took {end} seconds")
        return results

    def find_one(self,config,cols=None,dropna=False):
        return self.filter(config,cols,dropna).sample(frac=1).iloc[0]
        
    def find(self,config,cols=None,dropna=False):
        return self.filter(config,cols,dropna)

    def _filter(self,df,filters):
        ops = {
            '>': operator.gt,
            '<': operator.lt,
            '>=': operator.ge,
            '<=': operator.le,
            '==': operator.eq
        }


        if not isinstance(filters,dict):
            raise NotImplementedError("filter must be a 'dict' .")

        for col,value in filters.items():
            if isinstance(value,dict):
                for op_str,val in value.items():
                    df = df[ops[op_str](df[col],val)]                        
            else:
                df = df[df[col] == value]
        return df

    
    def filter(self,config,cols=None,dropna=False):
        retval = copy.deepcopy(self)
        for table,spec in config.items():
            df = retval[table]
            if isinstance(df,DestinationTable):
                df = df.get_df()
            for col,func in spec.items():
                df = df[df[col].apply(func)]

            retval[table] = df
            
        if dropna:
            retval = retval.dropna(axis=1)
        if cols is not None:
            index = retval.index.name
            retval = retval.reset_index()
            retval = retval[[col for col,keep in cols.items() if keep]]
            if index in retval.columns:
                retval = retval.set_index(index)
            
        return retval

        
        # for obj in config:
        #     if isinstance(obj,str):
        #         df = self[obj].get_df()
        #         if df.index.name != 'person_id':
        #             df = df.set_index('person_id')
        #     elif isinstance(obj,dict):
        #         for key,value in obj.items():
        #             print (self[key])
        #             df = self[key]
        #             df = self._filter(df,value)
        #             if df.index.name != 'person_id':
        #                 df = df.set_index('person_id')
        #             if retval is None:
        #                 retval = df
        #             else:
        #                 retval = retval.merge(df,left_index=True,right_index=True)
        #     else:
        #         raise NotImplementedError("need to pass a json object to filter()")

        
    def get_all_objects(self):
        return [ obj for collection in self.__objects.values() for obj in collection.values()]

    
    def get_start_index(self,destination_table):
        self.logger.debug(f'getting start index for {destination_table}')

        if self.indexing_conf == None or not self.indexing_conf :
            self.logger.debug(f"no indexing specified, so starting the index for {destination_table} from 1")
            return 1

        if destination_table in self.indexing_conf:
            return int(self.indexing_conf[destination_table])
        else:
            self.logger.warning(self.indexing_conf)
            self.logger.warning("indexing configuration has be parsed "
                                f"but this table ({destination_table}) "
                                "has not be passed, so starting from 1")
            return 1

    def clear_objects(self,destination_table=None):
        if destination_table:
            self.__objects[destination_table].clear()
        else:
            for destination_table in self.__objects:
                self.__objects[destination_table].clear()
            
    def get_objects(self,destination_table=None):
        """
        For a given destination table:
        * Retrieve all associated objects that have been registered with the class

        Args:
            destination_table (str) : name of a destination (CDM) table e.g. "person"
        Returns:
            list : a list of destination table objects 
                   e.g. [<person_0>, <person_1>] 
                   which would be objects for male and female mapping
        """
        if destination_table == None:
            return self.__objects
        
        self.logger.debug(f"looking for {destination_table}")
        if destination_table not in self.__objects.keys():
            self.logger.error(f"Trying to obtain the table '{destination_table}', but cannot find any objects")
            raise Exception("Something wrong!")

        return [
            obj
            for obj in self.__objects[destination_table].values()
        ]

    
    def mask_person_id(self,df,destination_table):
        """
        Given a dataframe object, apply a masking map on the person_id, if one has been created
        Args:
            df (pandas.Dataframe) : input pandas dataframe
        Returns:
            pandas.Dataframe: modified dataframe with person id masked
        """

        if 'person_id' in df.columns:
            #if masker has not been defined, define it
            if destination_table == 'person':
                if self.person_id_masker is not None:
                    start_index = int(list(self.person_id_masker.values())[-1]) + 1
                    new = False
                else:
                    self.person_id_masker = {}
                    start_index = self.get_start_index(destination_table)
                    new = True

                person_id_masker = {}
                for i,x in enumerate(df['person_id'].unique()):
                    index = i+start_index
                    if x in self.person_id_masker:
                        existing_index = self.person_id_masker[x]
                        self.logger.error(f"'{x}' already found in the person_id_masker")
                        self.logger.error(f"'{existing_index}' assigned to this already")
                        self.logger.error(f"was trying to set '{index}'")
                        self.logger.error(f"Most likely cause is this is duplicate data!")
                        raise PersonExists('Duplicate person found!')
                    person_id_masker[x] = index

                self.person_id_masker.update(person_id_masker)

                if self.outputs:
                    dfp = pd.DataFrame(((s,t)
                                        for t,s in person_id_masker.items()),
                                       columns=['SOURCE_SUBJECT','TARGET_SUBJECT'])

                    mode = 'w' if new else 'a'
                    self.outputs.write(f"person_ids",dfp,mode)
                        
            #apply the masking
            if self.person_id_masker is None:
                raise Exception(f"Person ID masking cannot be performed on"
                                f" {destination_table} as no masker based on a person table has been defined!")

            nbefore = len(df['person_id'])
            df['person_id'] = df['person_id'].map(self.person_id_masker)

            self.logger.debug(f"Just masked person_id using integers")
            if destination_table != 'person':
                df.dropna(subset=['person_id'],inplace=True)
                nafter = len(df['person_id'])
                ndiff = nbefore - nafter
                df.attrs['valid_person_id'] = {'before':nbefore,'after':nafter}
                #if rows have been removed
                if ndiff>0:
                    self.logger.error("There are person_ids in this table that are not in the output person table!")
                    self.logger.error("Either they are not in the original data, or while creating the person table, ")
                    self.logger.error("studies have been removed due to lack of required fields, such as birthdate.")
                    self.logger.error(f"{nafter}/{nbefore} were good, {ndiff} studies are removed.")
            
        return df

    def count_objects(self):
        """
        For each CDM (destination) table, count the number of objects associated
        e.g.
        {
           "observation": 6,
           "condition_occurrence": 1,
           "person": 2
        }
        """
        count_map = json.dumps({
            key: len(obj.keys())
            for key,obj in self.__objects.items()
        },indent=6)
        self.logger.info(f"Number of objects to process for each table...\n{count_map}")

    def keys(self):
        """
        For cdm.keys(), return the keys of which objects have been mapped.
        Hence which CDM table dataframes have been created.
        This should be used AFTER cdm.process() has been run, which creates the dataframes.
        """
        return self.__df_map.keys()

    def objects(self):
        """
        Method to retrieve the input objects to the CDM 
        """
        return self.__objects

                    
    def process(self,object_list=None,conserve_memory=False):
        """
        Process chunked data, processes as follows
        * While the chunking of is not yet finished
        * Loop over all CDM tables (via execution order) 
          and process the table (see process_table), returning a dataframe
        * Register the retrieve dataframe with the model  
        * For the current chunk slice, save the data/logs to files
        * Retrieve the next chunk of data

        """
        self.execution_order = self.get_execution_order()
        self.logger.info(f"Starting processing in order: {self.execution_order}")
        self.count_objects()

        for destination_table in self.execution_order:
            first = True
            i = 0
            while True:                
                df_generator = self.process_table(destination_table,object_list=object_list)
                ntables = 0
                nrows = 0
                dfs = []
                #print (self.drop_duplicates)
                for j,obj in enumerate(df_generator):

                    df = obj.get_df()
                    ntables +=1
                    nrows += len(df)
                    if conserve_memory and self.save_files:
                        mode = None if first else 'a'
                        self.save_dataframe(destination_table,df,mode=mode)
                        first = False
                        obj.clear()
                        del df
                        df = None
                    else:
                        dfs.append(df)

                if not conserve_memory:
                    df = pd.concat(dfs,ignore_index=True)#.sort_values(df.columns[0])
                    if self.save_files:
                        if self.drop_duplicates and destination_table != 'person':
                            nbefore = len(df)
                            df_hash = pd.util.hash_pandas_object(df.drop(df.columns[0],axis=1),index=False)
                            df_temp = df[df_hash.duplicated(keep=False)].head(10).dropna(axis=1)
                            df = df[~df_hash.duplicated()]
                            nafter = len(df)
                            ndiff = nbefore - nafter
                            if ndiff>0:
                                self.logger.error(f"Removed {ndiff} row(s) due to duplicates found when merging {destination_table}")
                                self.logger.warning("Example duplicates...")
                                self.logger.warning(df_temp.set_index(df_temp.columns[0]))
                        
                        mode = None if first else 'a'
                        self.save_dataframe(destination_table,df,mode=mode)
                        first = False

                    for col in df.columns:
                        if col.endswith("_id"):
                            df[col] = df[col].astype(float).astype(pd.Int64Dtype())
                    if df.index.name == 'index' or df.index.name is None:
                        df = df.set_index(df.columns[0])
                    
                    self[destination_table] = df


                self.logger.info(f'finalised {destination_table} on iteration {i} producing {nrows} rows from {ntables} tables')
                
                #move onto the next iteration                
                i+=1

                if self.inputs:
                    try:
                        self.inputs.next()
                    except StopIteration:
                        break
                else:
                    break

            #if inputs are defined, and we havent just finished the last table,
            #reset the inputs
            if self.inputs and not destination_table == self.execution_order[-1]:
                self.inputs.reset()

        #for destination_table in self.execution_order:
        #    index = self.get_start_index(destination_table)
        #    print (index)

                
    def process_simult(self,object_list=None,conserve_memory=False):
        """
        process simulataneously
        """
        self.execution_order = self.get_execution_order()
        self.logger.info(f"Starting processing in order: {self.execution_order}")
        self.count_objects()
        i=0
        while True:
            for destination_table in self.execution_order:
                df_generator = self.process_table(destination_table,object_list=object_list)
                ntables = 0
                nrows = 0
                dfs = []
                for j,obj in enumerate(df_generator):
                    df = obj.get_df()
                        
                    ntables +=1
                    nrows += len(df)
                    if self.save_files:
                        mode = None if j==0 else 'a'
                        self.save_dataframe(destination_table,df,mode=mode)
                        first = False
                    if not conserve_memory:
                        dfs.append(df)
                    #else:
                    #    obj.clear()
                    #    del df
                    #    df = None
                if not conserve_memory:
                    self[destination_table] = pd.concat(dfs,ignore_index=True)
                        
            if self.inputs:
                try:
                    self.inputs.next()
                except StopIteration:
                    break
            else:
                break
            i+=1



    def get_tables(self):
        return list(self.__objects.keys())
    
    def get_execution_order(self):
        if not self.execution_order:
            self.execution_order = sorted(self.__objects.keys(), key=lambda x: x != 'person')
        return self.execution_order
    
    def set_execution_order(self,order):
        self.execution_order = order
            
    def process_table(self,destination_table,object_list=None):
        """
        Process a CDM (destination) table. The method proceeds as follows:
        * Given a destination table name e.g. 'person' 
        * Retrieve all objects belonging to the given CDM table (e.g. <person_0, person_1>)
        * Loop over each object
        * Retrieve a dataframe for that object given it's definition/rules (see get_df)
        * Concatenate all retrieve dataframes together by stacking on top of each other vertically
        * Create new indexes for the primary column so the go from 1-N 
        
        Args:
            destination_table (str) : name of a destination table to process (e.g. 'person')
            object_list (list) : [optional] list of objects to process
        Returns:
            list(pandas.Dataframe): a dataframes in the CDM format for this destination table
        """
        objects = self.get_objects(destination_table)
        if object_list:
            objects = [obj for obj in object_list if obj in objects]
        
        nobjects = len(objects)
        extra = ""
        if nobjects>1:
            extra="s"
        self.logger.info(f"for {destination_table}: found {nobjects} object{extra}")
        
        if len(objects) == 0:
            yield None
        
        #execute them all
        dfs = []
        self.logger.info(f"working on {destination_table}")
        logs = {'objects':{}}

        if destination_table not in self.logs['meta']['total_data_processed']:
            self.logs['meta']['total_data_processed'][destination_table] = 0

        nrows_processed = self.logs['meta']['total_data_processed'][destination_table]
            
        for i,obj in enumerate(objects):
            self.logger.info(f"starting on {obj.name}")

            start_index = self.get_start_index(destination_table)
            start_index += nrows_processed

            #force_rebuild=True,
            df = obj.get_df(start_index=start_index)

            self.logger.info(f"finished {obj.name} ({hex(id(df))}) "
                             f"... {i+1}/{len(objects)} completed, {len(df)} rows") 
            if len(df) == 0:
                self.logger.warning(f".. no outputs were found ")
                continue

            if self.do_mask_person_id:
                df = self.mask_person_id(df,destination_table)

            obj._meta.update(df.attrs)
            nrows_processed += len(df)
            self.logs['meta']['total_data_processed'][destination_table] = nrows_processed
            if destination_table not in self.logs:
                self.logs[destination_table] = {}
                
            self.logs[destination_table][hex(id(df))] = obj._meta
            
            obj.set_df(df)
            yield obj

    def save_dataframe(self,table,df=None,mode=None):
        if self.outputs:
            _id = hex(id(df))
            self.logger.info(f"saving dataframe ({_id}) to {self.outputs}")
            self.outputs.write(table,df,mode)
        else:
            self.logger.info(f"called save_dateframe but outputs are not defined. save_files: {self.save_files}")
            
    def set_person_id_map(self,person_id_map):
        self.person_id_masker = person_id_map

    def set_indexing_map(self,indexing):
        self.indexing_conf = indexing
        
    def set_outfile_separator(self,sep):
        """
        Set which separator to use, e.g. ',' or '\t' 

        Args:
            sep (str): which separator to use when writing csv (tsv) files
        """
        self._outfile_separator = sep
        
    def set_indexing(self,index_map,strict_check=False):
        """
        Create indexes on input files which would allow rules to use data from 
        different input tables.

        Args:
            index_map (dict): a map between the filename and what should be the column used for indexing 

        """
        if self.inputs == None:
            raise NoInputFiles('Trying to indexing before any inputs have been setup')

        for key,index in index_map.items():
            if key not in self.inputs.keys():
                self.logger.warning(f"trying to set index '{index}' for '{key}' but this has not been loaded as an inputs!")
                continue

            if index not in self.inputs[key].columns:
                self.logger.error(f"trying to set index '{index}' on dataset '{key}', but this index is not in the columns! something really wrong!")
                continue
            self.inputs[key].index = self.inputs[key][index].rename('index') 

